"""General functions for experiments and pytorch."""
import inspect
from pydoc import locate
from typing import Any, Optional

import matplotlib.pyplot as plt
import numpy as np
import torch

from sdfest.initialization import quaternion_utils


def str_to_object(name: str) -> Any:
    """Try to find object with a given name.

    First scope of calling function is checked for the name, then current environment
    (in which case name has to be a fully qualified name). In the second case, the
    object is imported if found.

    Args:
        name: Name of the object to resolve.
    Returns:
        The object which the provided name refers to. None if no object was found.
    """
    # check callers local variables
    caller_locals = inspect.currentframe().f_back.f_locals
    if name in caller_locals:
        return caller_locals[name]

    # check callers global variables (i.e., imported modules etc.)
    caller_globals = inspect.currentframe().f_back.f_globals
    if name in caller_globals:
        return caller_globals[name]

    # check environment
    return locate(name)


def visualize_sample(sample: Optional[dict] = None, prediction: Optional[dict] = None):
    """Visualize sample and prediction.

    Assumes the following conventions and keys
        "scale": Half maximum side length of bounding box.
        "quaternion: Scalar-last orientation of object.
    """
    pointset = sample["pointset"].cpu().numpy()
    plt.imshow(sample["mask"].cpu().numpy())
    plt.show()
    plt.imshow(sample["depth"].cpu().numpy())
    plt.show()
    fig = plt.figure()
    ax = fig.add_subplot(projection="3d")
    ax.set_box_aspect((1, 1, 1))
    max_points = 500
    if len(pointset) > max_points:
        indices = np.random.choice(len(pointset), replace=False, size=max_points)
        ax.scatter(pointset[indices, 0], pointset[indices, 1], pointset[indices, 2])
    else:
        ax.scatter(pointset[:, 0], pointset[:, 1], pointset[:, 2])

    _plot_coordinate_frame(ax, sample)

    _plot_bounding_box(ax, sample)

    set_axes_equal(ax)

    plt.show()


def _plot_coordinate_frame(ax, sample):
    axis_pts = sample["scale"].cpu() * torch.eye(3)
    axis_pts = quaternion_utils.quaternion_apply(sample["quaternion"].cpu(), axis_pts)
    axis_pts = axis_pts + sample["position"].cpu()
    axis_pts = axis_pts.numpy()
    origin = sample["position"].cpu()
    _plot_line(ax, origin, axis_pts[0], "r")
    _plot_line(ax, origin, axis_pts[1], "g")
    _plot_line(ax, origin, axis_pts[2], "b")


def _plot_bounding_box(ax, sample):
    border_pts = (
        torch.tensor(
            [
                [-1.0, -1.0, -1.0],
                [-1.0, -1.0, 1.0],
                [-1.0, 1.0, -1.0],
                [-1.0, 1.0, 1.0],
                [1.0, -1.0, -1.0],
                [1.0, -1.0, 1.0],
                [1.0, 1.0, -1.0],
                [1.0, 1.0, 1.0],
            ]
        )
        * sample["scale"].cpu()
    )
    border_pts = quaternion_utils.quaternion_apply(
        sample["quaternion"].cpu(), border_pts
    )
    border_pts = border_pts + sample["position"].cpu()
    _plot_line(ax, border_pts[0], border_pts[1], "k")
    _plot_line(ax, border_pts[0], border_pts[2], "k")
    _plot_line(ax, border_pts[0], border_pts[4], "k")
    _plot_line(ax, border_pts[1], border_pts[3], "k")
    _plot_line(ax, border_pts[1], border_pts[5], "k")
    _plot_line(ax, border_pts[2], border_pts[3], "k")
    _plot_line(ax, border_pts[2], border_pts[6], "k")
    _plot_line(ax, border_pts[3], border_pts[7], "k")
    _plot_line(ax, border_pts[4], border_pts[5], "k")
    _plot_line(ax, border_pts[4], border_pts[6], "k")
    _plot_line(ax, border_pts[5], border_pts[7], "k")
    _plot_line(ax, border_pts[6], border_pts[7], "k")


def _plot_line(ax, pt1, pt2, *args, **kwargs):
    ax.plot([pt1[0], pt2[0]], [pt1[1], pt2[1]], [pt1[2], pt2[2]], *args, **kwargs)


def set_axes_equal(ax) -> None:
    """Make axes of 3D plot have equal scale.

    This ensures that spheres appear as spheres, cubes as cubes, ...
    This is needed since Matplotlib's ax.set_aspect('equal') and
    and ax.axis('equal') are not supported for 3D.

    From: https://stackoverflow.com/a/31364297

    Args:
      ax: A Matplotlib axis, e.g., as output from plt.gca().
    """
    x_limits = ax.get_xlim3d()
    y_limits = ax.get_ylim3d()
    z_limits = ax.get_zlim3d()

    x_range = abs(x_limits[1] - x_limits[0])
    x_middle = np.mean(x_limits)
    y_range = abs(y_limits[1] - y_limits[0])
    y_middle = np.mean(y_limits)
    z_range = abs(z_limits[1] - z_limits[0])
    z_middle = np.mean(z_limits)

    # The plot bounding box is a sphere in the sense of the infinity
    # norm, hence I call half the max range the plot radius.
    plot_radius = 0.5 * max([x_range, y_range, z_range])

    ax.set_xlim3d([x_middle - plot_radius, x_middle + plot_radius])
    ax.set_ylim3d([y_middle - plot_radius, y_middle + plot_radius])
    ax.set_zlim3d([z_middle - plot_radius, z_middle + plot_radius])


def dict_to(data_dict: dict, device: torch.device) -> dict:
    """Move values in dictionary of type torch.Tensor to a specfied device.

    Args:
        data_dict: Dictionary to be iterated over.
        device: Device to move objects of type torch.Tensor to.
    Returns:
        Dictionary containing the same keys and values as data_dict, but with all
        objects of type torch.Tensor moved to the specified device.
    """
    new_data_dict = {}
    for k, v in data_dict.items():
        if isinstance(v, torch.Tensor):
            new_data_dict[k] = v.to(device)
        else:
            new_data_dict[k] = v
    return new_data_dict



"""Experiments with differentiable SDF renderer.

Usage:
    python -m sdfest.differentiable_renderer.scripts.experiments --sdf_path data/shapenet_processed/mug_filtered/00001.npy --scale 0.15  \
    --scale_off --pos 0 0 -0.4 --pos_off 0.03 0.03 0.03 --rot 400 90 0 --rot_off 10 10 10 --gpu --visualize 10 --steps 1000
"""
import torch
import time
import click
import numpy as np
import matplotlib.pyplot as plt
from scipy.spatial.transform import Rotation

from sdfest.differentiable_renderer.sdf_renderer import render_depth, render_depth_gpu


@click.command()
@click.option("--sdf_path", type=click.Path(exists=True), required=True)
@click.option("--ref_sdf_path", type=click.Path(exists=True), default=None)
@click.option("--pos", type=float, nargs=3, default=None)
@click.option("--rot", type=float, nargs=3, default=None)
@click.option("--scale", type=float, default=None)
@click.option("--pos_off", type=float, nargs=3, default=None)
@click.option("--rot_off", type=float, nargs=3, default=None)
@click.option("--scale_off", type=float, default=None)
@click.option("--steps", type=int, default=100)
@click.option("--width", type=int, default=640)
@click.option("--height", type=int, default=480)
@click.option("--fov", type=float, default=90)
@click.option("--gpu", is_flag=True)
@click.option("--visualize", type=int, default=None)
@click.option("--threshold", type=float, default=0.01)
def offset_experiment(
    sdf_path,
    ref_sdf_path,
    pos,
    rot,
    scale,
    pos_off,
    rot_off,
    scale_off,
    steps,
    width,
    height,
    fov,
    gpu,
    visualize,
    threshold,
):
    # set device
    if gpu and not torch.cuda.is_available():
        gpu = False
    device = torch.device("cuda" if gpu else "cpu")

    if pos is None or not pos:
        pos = [0.0, 0.0, 0.0]
    if rot is None or not rot:
        rot = [0.0, 0.0, 0.0]
    if pos_off is None or not pos_off:
        pos_off = [0.0, 0.0, 0.0]
    if rot_off is None or not rot_off:
        rot_off = [0.0, 0.0, 0.0]
    if ref_sdf_path is None:
        ref_sdf_path = sdf_path

    ref_pos, ref_rot = torch.tensor(pos, device=device), torch.tensor(
        rot, device=device
    )
    ref_scale_inv = torch.tensor(1.0 / scale, device=device)
    pos_off, rot_off = torch.tensor(pos_off, device=device), torch.tensor(
        rot_off, device=device
    )
    pos = ref_pos + pos_off
    rot = ref_rot + rot_off
    scale_inv = torch.tensor(1.0 / (scale + scale_off), device=device)
    sdf = torch.from_numpy(np.load(sdf_path)).float().to(device)
    ref_sdf = torch.from_numpy(np.load(ref_sdf_path)).float().to(device)
    quat = (
        torch.from_numpy(Rotation.from_euler("XYZ", rot.cpu(), True).as_quat())
        .float()
        .to(device)
    )
    ref_quat = (
        torch.from_numpy(Rotation.from_euler("XYZ", ref_rot.cpu(), True).as_quat())
        .float()
        .to(device)
    )

    # Enable gradients for all optimizable variables
    quat.requires_grad_()
    pos.requires_grad_()
    sdf.requires_grad_()
    scale_inv.requires_grad_()

    _, ((ax1, ax2), (ax3, ax4)) = plt.subplots(2, 2)
    plt.ion()
    plt.show()

    # Generate reference image
    with torch.no_grad():
        if gpu:
            t1 = time.time()
            ref = render_depth_gpu(
                ref_sdf, ref_pos, ref_quat, ref_scale_inv, width, height, fov, threshold
            )
            # ref[:,0:250] = 0
            # ref[:,200:] = 0
        else:
            t1 = time.time()
            ref = render_depth(
                ref_sdf, ref_pos, ref_quat, ref_scale_inv, width, height, fov, threshold
            )

    optimizer = torch.optim.Adam([pos, quat, scale_inv], lr=0.001)

    # run the optimization
    step = 0
    total_time = 0
    while step < steps:
        t1 = time.time()
        if gpu:
            img = render_depth_gpu(
                sdf, pos, quat, scale_inv, width, height, fov, threshold
            )
        else:
            img = render_depth(sdf, pos, quat, scale_inv, width, height, fov, threshold)
        mask = (img > 0) & (ref > 0)
        error = (img - ref) ** 2
        loss = torch.mean(torch.abs((img - ref)[mask]))
        loss.backward()
        with torch.no_grad():
            step += 1
            optimizer.step()
            optimizer.zero_grad()
            quat /= torch.sqrt(torch.sum(quat ** 2))  # normalize quaternion
            total_time += time.time() - t1
            print(f"loss: {loss} step: {step} s/step {total_time / step}")

            if visualize is not None and step % visualize == 0:
                ax1.clear()
                ax2.clear()
                ax3.clear()
                ax4.clear()
                ax1.imshow(img.detach().cpu().numpy(), cmap="gist_heat_r")
                ax2.imshow(img.detach().cpu().numpy(), cmap="gist_heat_r", alpha=0.5)
                ax2.imshow(ref.detach().cpu().numpy(), cmap="gist_heat_r", alpha=0.5)
                vmax = torch.max(error[mask]).detach().cpu().numpy()
                ax3.imshow(
                    error.detach().cpu().numpy(), cmap="inferno", vmin=0, vmax=vmax
                )
                ax4.imshow(ref.detach().cpu().numpy(), cmap="gist_heat_r")
                plt.pause(0.00001)


if __name__ == "__main__":
    offset_experiment()



"""Utility functions to handle various datasets."""
import random
from typing import Iterator, List

import numpy as np
import torch


def collate_samples(samples: List[dict]) -> dict:
    """Collate sample dictionaries.

    Performs standard batching and additionally batches pointsets by taking subset of
    points.
    Also supports non-tensor types, which will be returned as standard lists.

    Reduces all pointsets to a common size based on the smallest set.

    Args:
        samples:
            Dictionary containing various types of data.
            All keys except "pointset" will use standard batching.
            All samples are expected to contain the same keys.
    Returns:
        Dictionary containing same keys as each sample.
        For "pointset" key:
            Tensor of size (N, M_min, D) where N is the batch size, M_min the number of
            points in the smallest pointset and D the number of channels per point.
    """
    batch = {}

    for key in samples[0].keys():
        if key == "pointset":
            batch_size = len(samples)

            smallest_set = min(s["pointset"].shape[0] for s in samples)
            # limit number of points to limit memory usage
            smallest_set = min(smallest_set, 2500)

            sample_pointset = samples[0]["pointset"]

            channels = sample_pointset.shape[-1]
            device = sample_pointset.device
            batch["pointset"] = torch.empty(
                batch_size, smallest_set, channels, device=device
            )
            for i, sample in enumerate(samples):
                num_points = sample["pointset"].shape[0]
                point_indices = random.sample(range(0, num_points), smallest_set)
                batch["pointset"][i] = sample["pointset"][point_indices]
        elif isinstance(samples[0][key], torch.Tensor):
            # standard batching for torch tensors
            batch[key] = torch.stack([s[key] for s in samples])
        else:
            # standard list for other data types
            batch[key] = [s[key] for s in samples]

    return batch


class MultiDataLoader:
    """Wrapper for multiple dataloaders."""

    def __init__(
        self,
        data_loaders: List[torch.utils.data.DataLoader],
        probabilities: List[float],
    ) -> None:
        """Initialize the class."""
        self._data_loaders = data_loaders
        self._data_loader_iterators = [iter(dl) for dl in self._data_loaders]
        self._probabilities = probabilities
        assert len(self._data_loaders) == len(self._probabilities)

    def __iter__(self) -> Iterator:
        """Return infinite iterator which returns samples from sampled data_loader."""
        while True:
            i = np.random.choice(
                np.arange(len(self._probabilities)), p=self._probabilities
            )
            try:
                yield next(self._data_loader_iterators[i])
            except StopIteration:
                self._data_loader_iterators[i] = iter(self._data_loaders[i])



"""Script to train initialization model."""
import argparse
from collections import defaultdict
from datetime import datetime, timedelta
import os
import random
import time

import numpy as np
from sdfest.vae.sdf_vae import SDFVAE
import torch
import torchinfo
from tqdm import tqdm
import wandb
import yoco

import sdfest
from sdfest.initialization.datasets import dataset_utils
from sdfest.initialization.datasets.generated_dataset import SDFVAEViewDataset
from sdfest.initialization.sdf_pose_network import SDFPoseNet, SDFPoseHead
from sdfest.initialization.pointnet import VanillaPointNet
from sdfest.initialization import quaternion_utils, sdf_utils, utils
from sdfest.utils import load_model_weights

os.environ["PYOPENGL_PLATFORM"] = "egl"

MODULE_DICT = {c.__name__: c for c in [SDFPoseHead, VanillaPointNet]}


class Trainer:
    """Trainer for single shot pose and shape estimation network."""

    def __init__(self, config: dict) -> None:
        """Construct trainer.

        Args:
            config: The configuration for model and training.
        """
        self._read_config(config)

    def _read_config(self, config: dict) -> None:
        self._config = config
        self._validation_iteration = config["validation_iteration"]
        self._visualization_iteration = config["visualization_iteration"]
        self._checkpoint_iteration = config["checkpoint_iteration"]
        self._iterations = config["iterations"]
        self._init_weights_path = (
            config["init_weights"] if "init_weights" in config else None
        )

        # propagate orientation representation and category to datasets
        datasets = list(self._config["datasets"].values()) + list(
            self._config["validation_datasets"].values()
        )
        for dataset in datasets:
            dataset["config_dict"]["orientation_repr"] = config["orientation_repr"]
            if "orientation_grid_resolution" in config:
                dataset["config_dict"]["orientation_grid_resolution"] = config[
                    "orientation_grid_resolution"
                ]
            if "category_str" in config:
                dataset["config_dict"]["category_str"] = config["category_str"]

        # propagate orientation representation to init head
        self._config["head"]["orientation_repr"] = config["orientation_repr"]
        if "orientation_grid_resolution" in config:
            self._config["head"]["orientation_grid_resolution"] = config[
                "orientation_grid_resolution"
            ]

    def run(self) -> None:
        """Train the model."""
        wandb.init(project="sdfest.initialization", config=self._config)

        self._device = self.get_device()

        # init dataset and dataloader
        self.vae = self.create_sdfvae()

        # init model to train
        self._sdf_pose_net = SDFPoseNet(
            backbone=MODULE_DICT[self._config["backbone_type"]](
                **self._config["backbone"]
            ),
            head=MODULE_DICT[self._config["head_type"]](
                shape_dimension=self._config["vae"]["latent_size"],
                **self._config["head"],
            ),
        ).to(self._device)
        self._sdf_pose_net.train()

        # deterministic samples (needs to be done after model initialization, as it
        # can have varying number of parameters)
        torch.manual_seed(0)
        random.seed(torch.initial_seed())  # to get deterministic examples

        # print network summary
        torchinfo.summary(self._sdf_pose_net, (1, 500, 3), device=self._device)

        # init optimizer
        self._optimizer = torch.optim.Adam(
            self._sdf_pose_net.parameters(), lr=self._config["learning_rate"]
        )

        # load weights if provided
        if self._init_weights_path is not None:
            state_dict = torch.load(self._init_weights_path, map_location=self._device)
            self._sdf_pose_net.load_state_dict(state_dict)

        self._current_iteration = 0
        self._run_name = (
            f"sdfest.initialization_{datetime.now().strftime('%Y-%m-%d_%H-%M-%S-%f')}"
        )

        wandb.config.run_name = (
            self._run_name  # to allow association of pt files with wandb runs
        )

        self._model_base_path = os.path.join(os.getcwd(), "models", self._run_name)

        self._multi_data_loader = self._create_multi_data_loader()
        self._validation_data_loader_dict = self._create_validation_data_loader_dict()

        # backup config to model directory
        os.makedirs(self._model_base_path, exist_ok=True)
        config_path = os.path.join(self._model_base_path, "config.yaml")
        yoco.save_config_to_file(config_path, self._config)

        self._start_time = time.time()
        for samples in self._multi_data_loader:
            self._current_iteration += 1
            print(f"Current iteration: {self._current_iteration}\033[K", end="\r")
            self._update_progress()

            samples = utils.dict_to(samples, self._device)

            latent_shape, position, scale, orientation = self._sdf_pose_net(
                samples["pointset"]
            )
            predictions = {
                "latent_shape": latent_shape,
                "position": position,
                "scale": scale,
                "orientation": orientation,
            }

            loss = self._compute_loss(samples, predictions)
            self._optimizer.zero_grad()
            loss.backward()
            self._optimizer.step()

            with torch.no_grad():
                # samples_dict = defaultdict(lambda: dict())
                # for k,vs in samples.items():
                #     for i, v in enumerate(vs):
                #         samples_dict[i][k] = v
                # for sample in samples_dict.values():
                #     utils.visualize_sample(sample, None)

                self._compute_metrics(samples, predictions)

                if self._current_iteration % self._visualization_iteration == 0:
                    self._generate_visualizations()

                if self._current_iteration % self._validation_iteration == 0:
                    self._compute_validation_metrics()

            if self._current_iteration % self._checkpoint_iteration == 0:
                self._save_checkpoint()

            if self._current_iteration >= self._iterations:
                break

        now = time.time()
        print(f"Training finished after {now-self._start_time} seconds.")

        # save the final model
        torch.save(
            self._sdf_pose_net.state_dict(),
            os.path.join(wandb.run.dir, f"{wandb.run.name}.pt"),
        )
        config_path = os.path.join(wandb.run.dir, f"{wandb.run.name}.yaml")
        self._config["model"] = os.path.join(".", f"{wandb.run.name}.pt")
        yoco.save_config_to_file(config_path, self._config)

    def get_device(self) -> torch.device:
        """Create device based on config."""
        if "device" not in self._config or self._config["device"] is None:
            return torch.device("cuda" if torch.cuda.is_available() else "cpu")
        return torch.device(self._config["device"])

    def create_sdfvae(self) -> SDFVAE:
        """Create SDFVAE based on config.

        Returns:
            The SDFVAE on the specified device, with weights from specified model.
        """
        model_url = self._config["vae"].get("model_url")
        device = self.get_device()
        vae = SDFVAE(
            sdf_size=64,
            latent_size=self._config["vae"]["latent_size"],
            encoder_dict=self._config["vae"]["encoder"],
            decoder_dict=self._config["vae"]["decoder"],
            device=device,
        ).to(device)
        load_model_weights(self._config["vae"]["model"], vae, device, model_url)
        vae.eval()
        return vae

    def _compute_loss(
        self,
        samples: dict,
        predictions: dict,
    ) -> torch.Tensor:
        """Compute total loss.

        Args:
            samples:
                Samples dictionary containing a subset of the following keys:
                    "latent_shape": Shape (N,D).
                    "position": Shape (N,3).
                    "scale": Shape (N,).
                    "orientation":
                        Shape (N,4) for quaternion representation.
                        Shape (N,) for discretized representation.
            predictions: Dictionary containing the following keys:
                "latent_shape": Shape (N,D).
                "position": Shape (N,3).
                "scale": Shape (N,).
                "orientation":
                    Shape (N,4) for quaternion representation.
                    Shape (N,R) for discretized representation.
        Returns:
            The combined loss. Scalar.
        """
        log_dict = {}

        loss = 0

        if "latent_shape" in samples:
            loss_latent_l2 = torch.nn.functional.mse_loss(
                predictions["latent_shape"], samples["latent_shape"]
            )
            log_dict["loss latent"] = loss_latent_l2.item()
            loss = loss + self._config["latent_weight"] * loss_latent_l2

        if "position" in samples:
            loss_position_l2 = torch.nn.functional.mse_loss(
                predictions["position"], samples["position"]
            )
            log_dict["loss position"] = loss_position_l2.item()
            loss = loss + self._config["position_weight"] * loss_position_l2

        if "scale" in samples:
            loss_scale_l2 = torch.nn.functional.mse_loss(
                predictions["scale"], samples["scale"]
            )
            log_dict["loss scale"] = loss_scale_l2.item()
            loss = loss + self._config["scale_weight"] * loss_scale_l2

        if "orientation" in samples:
            if self._config["head"]["orientation_repr"] == "quaternion":
                loss_orientation = quaternion_utils.simple_quaternion_loss(
                    predictions["orientation"], samples["orientation"]
                )
            elif self._config["head"]["orientation_repr"] == "discretized":
                loss_orientation = torch.nn.functional.cross_entropy(
                    predictions["orientation"], samples["orientation"]
                )
            else:
                raise NotImplementedError(
                    "Orientation repr "
                    f"{self._config['head']['orientation_repr']}"
                    " not supported."
                )
            log_dict["loss orientation"] = loss_orientation.item()
            loss = loss + self._config["orientation_weight"] * loss_orientation

        log_dict["total loss"] = loss.item()

        wandb.log(
            log_dict,
            step=self._current_iteration,
        )

        return loss

    def _create_multi_data_loader(self) -> dataset_utils.MultiDataLoader:
        data_loaders = []
        probabilities = []
        for dataset_dict in self._config["datasets"].values():
            if dataset_dict["probability"] == 0.0:
                continue
            dataset = self._create_dataset(
                dataset_dict["type"], dataset_dict["config_dict"]
            )
            num_workers = 8 if dataset_dict["type"] != "SDFVAEViewDataset" else 0
            shuffle = (
                False if isinstance(dataset, torch.utils.data.IterableDataset) else True
            )
            probabilities.append(dataset_dict["probability"])
            data_loader = torch.utils.data.DataLoader(
                dataset=dataset,
                batch_size=self._config["batch_size"],
                collate_fn=dataset_utils.collate_samples,
                drop_last=True,
                shuffle=shuffle,
                num_workers=num_workers,
            )
            data_loaders.append(data_loader)
        return dataset_utils.MultiDataLoader(data_loaders, probabilities)

    def _create_validation_data_loader_dict(self) -> dict:
        data_loader_dict = {}
        for dataset_name, dataset_dict in self._config["validation_datasets"].items():
            dataset = self._create_dataset(
                dataset_dict["type"], dataset_dict["config_dict"]
            )
            data_loader = torch.utils.data.DataLoader(
                dataset=dataset,
                batch_size=self._config["batch_size"],
                collate_fn=dataset_utils.collate_samples,
                num_workers=12,
            )
            data_loader_dict[dataset_name] = data_loader
        return data_loader_dict

    def _create_dataset(
        self, type_str: str, config_dict: dict
    ) -> torch.utils.data.Dataset:
        dataset_type = utils.str_to_object(type_str)
        if dataset_type == SDFVAEViewDataset:
            dataset = dataset_type(
                config=config_dict,
                vae=self.vae,
            )
        elif dataset_type is not None:
            dataset = dataset_type(config=config_dict)
        else:
            raise NotImplementedError(f"Dataset type {type_str} not supported.")
        return dataset

    def _mean_geodesic_distance(self, samples: dict, predictions: dict) -> torch.Tensor:
        target_quaternions = samples["quaternion"]
        if self._config["head"]["orientation_repr"] == "quaternion":
            predicted_quaternions = predictions["orientation"]
        elif self._config["head"]["orientation_repr"] == "discretized":
            predicted_quaternions = torch.empty_like(target_quaternions)
            for i, v in enumerate(predictions["orientation"]):
                index = v.argmax().item()
                quat = self._sdf_pose_net._head._grid.index_to_quat(index)
                predicted_quaternions[i, :] = torch.tensor(quat)
        else:
            raise NotImplementedError(
                "Orientation representation "
                f"{self._config['head']['orientation_repr']}"
                " is not supported"
            )
        geodesic_distances = quaternion_utils.geodesic_distance(
            target_quaternions, predicted_quaternions
        )
        return torch.mean(geodesic_distances)

    def _compute_metrics(self, samples: dict, predictions: dict) -> None:
        # compute metrics / i.e., loss and representation independent metrics
        # extract quaternion from orientation representation
        geodesic_distance = self._mean_geodesic_distance(samples, predictions)
        wandb.log(
            {
                "metric geodesic distance": geodesic_distance.item(),
            },
            step=self._current_iteration,
        )

    def _generate_visualizations(self) -> None:
        # generate visualizations
        if self._current_iteration % self._visualization_iteration == 0:
            # generate unseen input and target
            samples = next(iter(self._multi_data_loader))
            samples = utils.dict_to(samples, self._device)
            predictions = self._sdf_pose_net(samples["pointset"])
            input_pointcloud = samples["pointset"][0].detach().cpu().numpy()
            input_pointcloud = np.hstack(
                (
                    input_pointcloud,
                    np.full((input_pointcloud.shape[0], 1), 0),
                )
            )
            output_sdfs = self.vae.decode(predictions[0])
            output_sdf = output_sdfs[0][0].detach().cpu().numpy()
            output_position = predictions[1][0].detach().cpu().numpy()
            output_scale = predictions[2][0].detach().cpu().numpy()
            if self._config["head"]["orientation_repr"] == "quaternion":
                output_quaternion = predictions[3][0].detach().cpu().numpy()
            elif self._config["head"]["orientation_repr"] == "discretized":
                index = predictions[3][0].argmax().item()
                output_quaternion = self._sdf_pose_net._head._grid.index_to_quat(index)
            else:
                raise NotImplementedError(
                    "Orientation representation "
                    f"{self._config['head']['orientation_repr']}"
                    " is not supported"
                )
            output_pointcloud = sdf_utils.sdf_to_pointcloud(
                output_sdf, output_position, output_quaternion, output_scale
            )
            output_pointcloud = np.hstack(
                (
                    output_pointcloud,
                    np.full((output_pointcloud.shape[0], 1), 1),
                )
            )
            pointcloud = np.vstack((input_pointcloud, output_pointcloud))

            wandb.log(
                {"point_cloud": wandb.Object3D(pointcloud)},
                step=self._current_iteration,
            )

            output_pointcloud = sdf_utils.sdf_to_pointcloud(
                output_sdf,
                samples["position"][0].detach().cpu().numpy(),
                samples["quaternion"][0].detach().cpu().numpy(),
                samples["scale"][0].detach().cpu().numpy(),
            )
            output_pointcloud = np.hstack(
                (
                    output_pointcloud,
                    np.full((output_pointcloud.shape[0], 1), 1),
                )
            )
            pointcloud = np.vstack((input_pointcloud, output_pointcloud))
            wandb.log(
                {"point_cloud gt pose": wandb.Object3D(pointcloud)},
                step=self._current_iteration,
            )

    def _compute_validation_metrics(self) -> None:
        self._sdf_pose_net.eval()
        for name, data_loader in self._validation_data_loader_dict.items():
            metrics_dict = defaultdict(lambda: 0)
            sample_count = 0
            for samples in tqdm(data_loader, desc="Validation"):
                batch_size = samples["position"].shape[0]
                samples = utils.dict_to(samples, self._device)
                latent_shape, position, scale, orientation = self._sdf_pose_net(
                    samples["pointset"]
                )
                predictions = {
                    "latent_shape": latent_shape,
                    "position": position,
                    "scale": scale,
                    "orientation": orientation,
                }
                euclidean_distance = torch.linalg.norm(
                    predictions["position"] - samples["position"], dim=1
                )
                metrics_dict[f"{name} validation mean position error / m"] += torch.sum(
                    euclidean_distance
                ).item()
                metrics_dict[f"{name} validation mean scale error / m"] += torch.sum(
                    torch.abs(predictions["scale"] - samples["scale"])
                ).item()
                metrics_dict[f"{name} validation mean geodesic_distance / rad"] += (
                    self._mean_geodesic_distance(samples, predictions).item()
                    * batch_size
                )
                sample_count += batch_size
                if self._config["head"]["orientation_repr"] == "discretized":
                    metrics_dict[
                        f"{name} validation orientation mean NLL"
                    ] += torch.nn.functional.cross_entropy(
                        predictions["orientation"],
                        samples["orientation"],
                        reduction="sum",
                    )
            for metric_name in metrics_dict:
                metrics_dict[metric_name] /= sample_count
            wandb.log(metrics_dict, step=self._current_iteration)
        self._sdf_pose_net.train()

    def _save_checkpoint(self) -> None:
        checkpoint_path = os.path.join(
            self._model_base_path, f"{self._current_iteration}.pt"
        )
        torch.save(
            self._sdf_pose_net.state_dict(),
            checkpoint_path,
        )

    def _update_progress(self) -> None:
        current_time = time.time()
        duration = current_time - self._start_time
        iterations_per_sec = self._current_iteration / duration
        if self._current_iteration > 10:
            remaining_iterations = self._iterations - self._current_iteration
            remaining_secs = remaining_iterations / iterations_per_sec
            remaining_time_str = str(timedelta(seconds=round(remaining_secs)))
        else:
            remaining_time_str = "N/A"
        print(
            f"Current iteration: {self._current_iteration:>10} / {self._iterations}"
            f" {self._current_iteration / self._iterations * 100:>6.2f}%"
            f" Remaining time: {remaining_time_str}"  # remaining time
            "\033[K",  # clear until end of line
            end="\r",  # overwrite previous
        )


def main() -> None:
    """Entry point of the program."""
    # define the arguments
    parser = argparse.ArgumentParser(description="Training script for init network.")
    parser.add_argument("--config", default="configs/default.yaml", nargs="+")
    config = yoco.load_config_from_args(
        parser, search_paths=[".", "~/.sdfest/", sdfest.__path__[0]]
    )
    trainer = Trainer(config)
    trainer.run()


if __name__ == "__main__":
    main()



"""Metrics for shape evaluation."""
from typing import Optional

import numpy as np
import scipy.spatial
from scipy.spatial.transform import Rotation


def correct_thresh(
    position_gt: np.ndarray,
    position_prediction: np.ndarray,
    orientation_gt: Rotation,
    orientation_prediction: Rotation,
    extent_gt: Optional[np.ndarray] = None,
    extent_prediction: Optional[np.ndarray] = None,
    points_gt: Optional[np.ndarray] = None,
    points_prediction: Optional[np.ndarray] = None,
    position_threshold: Optional[float] = None,
    degree_threshold: Optional[float] = None,
    iou_3d_threshold: Optional[float] = None,
    fscore_threshold: Optional[float] = None,
    rotational_symmetry_axis: Optional[int] = None,
) -> int:
    """Classify a pose prediction as correct or incorrect.

    Args:
        position_gt: ground truth position, expected shape (3,).
        position_prediction: predicted position, expected shape (3,).
        position_threshold: position threshold in meters, no threshold if None
        orientation_q_qt: ground truth orientation, scalar-last quaternion, shape (4,)
        orientation_q_prediction:
            predicted orientation, scalar-last quaternion, shape (4,)
        extent_gt:
            bounding box extents, shape (3,)
            only used if IoU threshold specified
        extent_prediction:
            bounding box extents, shape (3,)
            only used if IoU threshold specified
        point_gt: set of true points, expected shape (N,3)
        points_rec: set of reconstructed points, expected shape (M,3)
        degree_threshold: orientation threshold in degrees, no threshold if None
        iou_3d_threshold: 3D IoU threshold, no threshold if None
        rotational_symmetry_axis:
            Specify axis along which rotation is ignored. If None, no axis is ignored.
            0 for x-axis, 1 for y-axis, 2 for z-axis.
    Returns:
        1 if error is below all provided thresholds.  0 if error is above one provided
        threshold.
    """
    if position_threshold is not None:
        position_error = np.linalg.norm(position_gt - position_prediction)
        if position_error > position_threshold:
            return 0
    if degree_threshold is not None:
        rad_threshold = degree_threshold * np.pi / 180.0
        if rotational_symmetry_axis is not None:
            p = np.array([0.0, 0.0, 0.0])
            p[rotational_symmetry_axis] = 1.0
            p1 = orientation_gt.apply(p)
            p2 = orientation_prediction.apply(p)
            rad_error = np.arccos(p1 @ p2)
        else:
            rad_error = (orientation_gt * orientation_prediction.inv()).magnitude()
        if rad_error > rad_threshold:
            return 0
    if iou_3d_threshold is not None:
        raise NotImplementedError("3D IoU is not impemented yet.")
        # TODO implement 3D IoU
        # starting point for proper implementation: https://github.com/google-research-datasets/Objectron/blob/c06a65165a18396e1e00091981fd1652875c97b5/objectron/dataset/iou.py#L6
        pass
    if fscore_threshold is not None:
        fscore = reconstruction_fscore(points_gt, points_prediction, 0.01)
        if fscore < fscore_threshold:
            return 0
    return 1


def mean_accuracy(
    points_gt: np.ndarray,
    points_rec: np.ndarray,
    p_norm: int = 2,
    normalize: bool = False,
) -> float:
    """Compute accuracy metric.

    Accuracy metric is the same as asymmetric chamfer distance from rec to gt.

    See, for example, Occupancy Networks Learning 3D Reconstruction in Function Space,
    Mescheder et al., 2019.

    Args:
        points_gt: set of true points, expected shape (N,3)
        points_rec: set of reconstructed points, expected shape (M,3)
        p_norm: which Minkowski p-norm is used for distance and nearest neighbor query
        normalize: whether to divide result by Euclidean extent of points_gt
    Returns:
        Arithmetic mean of p-norm from reconstructed points to closest (in p-norm)
        ground truth points.
    """
    kd_tree = scipy.spatial.KDTree(points_gt)
    d, _ = kd_tree.query(points_rec, p=p_norm)
    if normalize:
        return np.mean(d) / extent(points_gt)
    else:
        return np.mean(d)


def mean_completeness(
    points_gt: np.ndarray,
    points_rec: np.ndarray,
    p_norm: int = 2,
    normalize: bool = False,
) -> float:
    """Compute completeness metric.

    Completeness metric is the same as asymmetric chamfer distance from gt to rec.

    See, for example, Occupancy Networks Learning 3D Reconstruction in Function Space,
    Mescheder et al., 2019.

    Args:
        points_gt: set of true points, expected shape (N,3)
        points_rec: set of reconstructed points, expected shape (M,3)
        p_norm: which Minkowski p-norm is used for distance and nearest neighbor query
        normalize: whether to divide result by Euclidean extent of points_gt
    Returns:
        Arithmetic mean of p-norm from ground truth points to closest (in p-norm)
        reconstructed points.
    """
    kd_tree = scipy.spatial.KDTree(points_rec)
    d, _ = kd_tree.query(points_gt, p=p_norm)
    if normalize:
        return np.mean(d) / extent(points_gt)
    else:
        return np.mean(d)


def symmetric_chamfer(
    points_gt: np.ndarray,
    points_rec: np.ndarray,
    p_norm: int = 2,
    normalize: bool = False,
) -> float:
    """Compute symmetric chamfer distance.

    There are various slightly different definitions for the chamfer distance.

    Note that completeness and accuracy are themselves sometimes referred to as
    chamfer distances, with symmetric chamfer distance being the combination of the two.

    Chamfer L1 in the literature (see, for example, Occupancy Networks Learning 3D
    Reconstruction in Function Space, Mescheder et al., 2019) refers to using
    arithmetic mean (note that this is actually differently scaled from L1) when
    combining accuracy and completeness.

    Args:
        points_gt: set of true points, expected shape (N,3)
        points_rec: set of reconstructed points, expected shape (M,3)
        p_norm: which Minkowski p-norm is used for distance and nearest neighbor query
        normalize: whether to divide result by Euclidean extent of points_gt
    Returns:
        Arithmetic mean of accuracy and completeness metrics using the specified p-norm.
    """
    return (
        mean_completeness(points_gt, points_rec, p_norm=p_norm, normalize=normalize)
        + mean_accuracy(points_gt, points_rec, p_norm=p_norm, normalize=normalize)
    ) / 2


def completeness_thresh(
    points_gt: np.ndarray,
    points_rec: np.ndarray,
    threshold: float,
    p_norm: int = 2,
    normalize: bool = False,
) -> float:
    """Compute thresholded completion metric.

    See FroDO: From Detections to 3D Objects, Rünz et al., 2020.

    Args:
        points_gt: set of true points, expected shape (N,3)
        points_rec: set of reconstructed points, expected shape (M,3)
        threshold: distance threshold to count a point as correct
        p_norm: which Minkowski p-norm is used for distance and nearest neighbor query
        normalize: whether to divide distances by Euclidean extent of points_gt
    Returns:
        Ratio of ground truth points with closest reconstructed point closer than
        threshold (in p-norm).
    """
    kd_tree = scipy.spatial.KDTree(points_rec)
    d, _ = kd_tree.query(points_gt, p=p_norm)
    if normalize:
        return np.sum(d / extent(points_gt) < threshold) / points_gt.shape[0]
    else:
        return np.sum(d < threshold) / points_gt.shape[0]


def accuracy_thresh(
    points_gt: np.ndarray,
    points_rec: np.ndarray,
    threshold: float,
    p_norm: int = 2,
    normalize: bool = False,
) -> float:
    """Compute thresholded accuracy metric.

    See FroDO: From Detections to 3D Objects, Rünz et al., 2020.

    Args:
        points_gt: set of true points, expected shape (N,3)
        points_rec: set of reconstructed points, expected shape (M,3)
        threshold: distance threshold to count a point as correct
        p_norm: which Minkowski p-norm is used for distance and nearest neighbor query
        normalize: whether to divide distances by Euclidean extent of points_gt
    Returns:
        Ratio of reconstructed points with closest ground truth point closer than
        threshold (in p-norm).
    """
    kd_tree = scipy.spatial.KDTree(points_gt)
    d, _ = kd_tree.query(points_rec, p=p_norm)
    if normalize:
        return np.sum(d / extent(points_gt) < threshold) / points_rec.shape[0]
    else:
        return np.sum(d < threshold) / points_rec.shape[0]


def reconstruction_fscore(
    points_gt: np.ndarray,
    points_rec: np.ndarray,
    threshold: float,
    p_norm: int = 2,
    normalize: bool = False,
) -> float:
    """Compute reconstruction fscore.

    See What Do Single-View 3D Reconstruction Networks Learn, Tatarchenko, 2019

    Args:
        points_gt: set of true points, expected shape (N,3)
        points_rec: set of reconstructed points, expected shape (M,3)
        threshold: distance threshold to count a point as correct
        p_norm: which Minkowski p-norm is used for distance and nearest neighbor query
        normalize: whether to divide distances by Euclidean extent of points_gt
    Returns:
        Harmonic mean of precision (thresholded accuracy) and recall (thresholded
        completeness).
    """
    recall = completeness_thresh(
        points_gt, points_rec, threshold, p_norm=p_norm, normalize=normalize
    )
    precision = accuracy_thresh(
        points_gt, points_rec, threshold, p_norm=p_norm, normalize=normalize
    )
    if recall < 1e-7 or precision < 1e-7:
        return 0
    return 2 / (1 / recall + 1 / precision)


def extent(points: np.ndarray) -> float:
    """Compute largest Euclidean distance between any two points.

    Args:
        points_gt: set of true
        p_norm: which Minkowski p-norm is used for distance and nearest neighbor query
    Returns:
        Ratio of reconstructed points with closest ground truth point closer than
        threshold (in p-norm).
    """
    try:
        hull = scipy.spatial.ConvexHull(points)
    except scipy.spatial.qhull.QhullError:
        # fallback to brute force distance matrix
        return np.max(scipy.spatial.distance_matrix(points, points))

    # this is wasteful, if too slow implement rotating caliper method
    return np.max(
        scipy.spatial.distance_matrix(points[hull.vertices], points[hull.vertices])
    )



"""This module provides various PyTorch modules for working with SDFs."""
from typing import Optional, Union
from pydoc import locate

import torch
import torch.nn as nn

from . import torch_utils


class SDFVAE(nn.Module):
    """Variational Autoencoder for Signed Distance Fields."""

    def __init__(
        self,
        sdf_size: int,
        latent_size: int,
        encoder_dict: dict,
        decoder_dict: dict,
        device: torch.device,
        tsdf: Optional[Union[bool, float]] = False,
    ):
        """Initialize the SDFVAE module.

        Args:
            sdf_size:       depth/width/length of the sdf
            latent_size:    dimensions of latent representation
            encoder_dict:
                Arguments passed to encoder constructor.
                See SDFEncoder.__init__ for details.
            decoder_dict:
                Arguments passed to decoder constructor.
                See SDFDecoder.__init__ for details.
            device:
                The device this model will work on.
                This is used for tensors created inside this model for sampling.
            tsdf:
                Value to truncate the SDF at. False for untruncated SDF.
                For the input this is only done in prepare_input, not in the forward
                pass. The output is truncated in the forward pass.
        """
        super().__init__()

        self.latent_size = latent_size

        self._device = device

        self.encoder = SDFEncoder(sdf_size, latent_size, tsdf=tsdf, **encoder_dict)
        self.decoder = SDFDecoder(sdf_size, latent_size, tsdf=tsdf, **decoder_dict)
        self.sdf_size = sdf_size
        self._tsdf = tsdf

    def forward(self, x, enforce_tsdf=False):
        z, means, log_var = self.encode(x)

        recon_x = self.decoder(z, enforce_tsdf)

        return recon_x, means, log_var, z

    def sample(self, n=1):
        z = torch.randn([n, self.latent_size]).to(self._device)
        return z

    def encode(self, x):
        means, log_var = self.encoder(x)

        std = torch.exp(0.5 * log_var)
        eps = torch.randn([x.shape[0], self.latent_size]).to(self._device)
        z = eps * std + means
        return z, means, log_var

    def inference(self, n=1, enforce_tsdf=False):
        z = self.sample(n)

        recon_x = self.decoder(z, enforce_tsdf)

        return recon_x, z

    def decode(self, z, enforce_tsdf=False):
        """Returns decoded SDF.

        Args:
            Batch of latent shapes. Shape (N, L).
        Returns:
            The decoded SDF. Shape (N, C, D, D, D).
        """
        return self.decoder(z, enforce_tsdf)

    def prepare_input(self, sdfs: torch.Tensor) -> None:
        """Convert batched SDFs to expected input format.

        This will transform inputs as defined by the decoder.
        See SDFEncoder.prepare_input for details.

        Will be done in place without gradients.

        Args:
            sdfs: Batched SDFs, expected shape (N,C,D,D,D).
        """
        self.encoder.prepare_input(sdfs)


class SDFEncoder(nn.Module):
    def __init__(self, volume_size, latent_size, layer_infos, tsdf=False):
        """Create SDFEncoder, i.e., define trainable layers.

        Args:
            volume_size:
                Input size D of the volume.
                (i.e., input tensor is expected to be Nx1xDxDxD)
            latent_size: Dimensionality of the latent representation.
            layers:
                Dictionaries defining the layers before the final output layers.
                Required fields:
                    - fully qualified type (str)
                    - args (dict): params passed to init of type
                These layers need to construct an 2D tensor (including batch dimension).
        """
        super().__init__()

        in_channels = 1

        # define layers
        layers = []
        for layer_info in layer_infos:
            t = locate(layer_info["type"])
            layers.append(t(**layer_info["args"]))
        self._features = torch.nn.Sequential(*layers)

        with torch.no_grad():
            output_size = self._features(
                torch.zeros(1, in_channels, volume_size, volume_size, volume_size)
            ).shape

        self.linear_means = nn.Linear(output_size[1], latent_size)
        self.linear_log_var = nn.Linear(output_size[1], latent_size)

        self._tsdf = tsdf

    def forward(self, x):
        """Forward pass of the module.

        Args:

        Returns:

        """
        out = self._features(x)

        means = self.linear_means(out)
        log_vars = self.linear_log_var(out)

        return means, log_vars

    def prepare_input(self, sdfs: torch.Tensor) -> None:
        """Convert batched SDFs to expected input format.

        This will truncate the SDFs based on tsdf argument passed to constructor.

        Will be done in place without gradients.

        Args:
            sdfs: Batched SDFs, expected shape (N,C,D,D,D).
        """
        if self._tsdf is not False:
            with torch.no_grad():
                sdfs.clamp_(-self._tsdf, self._tsdf)


class SDFDecoder(nn.Module):
    def __init__(
        self,
        volume_size: int,
        latent_size: int,
        fc_layers: list,
        conv_layers: list,
        tsdf: Optional[Union[bool, float]] = False,
    ):
        super().__init__()

        # sanity checks
        self.sanity_check(volume_size, fc_layers, conv_layers)

        # create layers and store params
        self._volume_size = volume_size

        in_size = latent_size
        self._fc_layers = torch.nn.ModuleList()
        for fc_layer in fc_layers:
            self._fc_layers.append(nn.Linear(in_size, fc_layer["out"]))
            in_size = fc_layer["out"]
        self._fc_info = fc_layers

        self._conv_layers = torch.nn.ModuleList()
        for conv_layer in conv_layers:
            self._conv_layers.append(
                nn.Conv3d(
                    conv_layer["in_channels"],
                    conv_layer["out_channels"],
                    conv_layer["kernel_size"],
                )
            )
        self._conv_info = conv_layers

        self._tsdf = tsdf

    def sanity_check(self, volume_size, fc_dicts, conv_dicts):
        assert fc_dicts[-1]["out"] == (
            conv_dicts[0]["in_channels"] * conv_dicts[0]["in_size"] ** 3
        )

        for i, conv_dict in enumerate(conv_dicts[:-1]):
            assert conv_dict["out_channels"] == conv_dicts[i + 1]["in_channels"]

        assert conv_dicts[-1]["out_channels"] == 1

    def forward(self, z, enforce_tsdf=False):
        """Decode latent vectors to SDFs.

        Args:
            z: Batch of latent vectors. Expected shape of (N,latent_size).
        """
        out = z
        for fc_layer in self._fc_layers:
            out = nn.functional.relu(fc_layer(out))

        out = out.view(
            -1,
            self._conv_info[0]["in_channels"],
            self._conv_info[0]["in_size"],
            self._conv_info[0]["in_size"],
            self._conv_info[0]["in_size"],
        )

        for info, layer in zip(self._conv_info, self._conv_layers):
            # interpolate to match next input
            if out.shape[2] != info["in_size"]:
                out = torch.nn.functional.interpolate(
                    out,
                    size=(info["in_size"], info["in_size"], info["in_size"]),
                    mode="trilinear",
                    align_corners=False,
                )
            out = layer(out)
            if info["relu"]:
                out = nn.functional.relu(out)

        if out.shape[2] != self._volume_size:
            out = torch.nn.functional.interpolate(
                out,
                size=(self._volume_size, self._volume_size, self._volume_size),
                mode="trilinear",
                align_corners=False,
            )

        if self._tsdf is not False and enforce_tsdf:
            out = out.clamp(-self._tsdf, self._tsdf)

        return out



"""Module providing dataset class for annotated Redwood dataset."""
import json
import os
from typing import TypedDict, Optional

from scipy.spatial.transform import Rotation
import numpy as np
import open3d as o3d
import torch
from PIL import Image
from sdfest.differentiable_renderer import Camera
import yoco

from sdfest.estimation import synthetic
from sdfest.initialization import pointset_utils, quaternion_utils, so3grid


class AnnotatedRedwoodDataset(torch.utils.data.Dataset):
    """Dataset class for annotated Redwood dataset.

    Data can be found here:
    http://redwood-data.org/3dscan/index.html

    Annotations are part SDFEst repo.

    Expected directory format:
        {root_dir}/{category_str}/rgbd/{sequence_id}/...
        {ann_dir}/{sequence_id}.obj
        {ann_dir}/annotations.json
    """

    num_categories = 3
    category_id_to_str = {
        0: "bottle",
        1: "bowl",
        2: "mug",
    }
    category_str_to_id = {v: k for k, v in category_id_to_str.items()}

    class Config(TypedDict, total=False):
        """Configuration dictionary for annoated Redwood dataset.

        Attributes:
            root_dir: See AnnotatedRedwoodDataset docstring.
            ann_dir: See AnnotatedRedwoodDataset docstring.
            mask_pointcloud: Whether the returned pointcloud will be masked.
            normalize_pointcloud:
                Whether the returned pointcloud and position will be normalized, such
                that pointcloud centroid is at the origin.
            scale_convention:
                Which scale is returned. The following strings are supported:
                    "diagonal":
                        Length of bounding box' diagonal. This is what NOCS uses.
                    "max": Maximum side length of bounding box.
                    "half_max": Half maximum side length of bounding box.
                    "full": Bounding box side lengths. Shape (3,).
            camera_convention:
                Which camera convention is used for position and orientation. One of:
                    "opengl": x right, y up, z back
                    "opencv": x right, y down, z forward
                Note that this does not influence how the dataset is processed, only the
                returned position and quaternion.
            orientation_repr:
                Which orientation representation is used. One of:
                    "quaternion"
                    "discretized"
            orientation_grid_resolution:
                Resolution of the orientation grid.
                Only used if orientation_repr is "discretized".
            remap_y_axis:
                If not None, the Redwood y-axis will be mapped to the provided axis.
                Resulting coordinate system will always be right-handed.
                This is typically the up-axis.
                Note that NOCS object models are NOT aligned the same as ShapeNetV2.
                To get ShapeNetV2 alignment: -y
                One of: "x", "y", "z", "-x", "-y", "-z"
            remap_x_axis:
                If not None, the original x-axis will be mapped to the provided axis.
                Resulting coordinate system will always be right-handed.
                Note that NOCS object models are NOT aligned the same as ShapeNetV2.
                To get ShapeNetV2 alignment: z
                One of: "x", "y", "z", "-x", "-y", "-z"
            category_str:
                If not None, only samples from the matching category will be returned.
                See AnnotatedRedwoodDataset.category_id_to_str for admissible category
                strings.
        """

        root_dir: str
        ann_dir: str
        split: str
        mask_pointcloud: bool
        normalize_pointcloud: bool
        scale_convention: str
        camera_convention: str
        orientation_repr: str
        orientation_grid_resolution: int
        remap_y_axis: Optional[str]
        remap_x_axis: Optional[str]
        category_str: Optional[str]

    default_config: Config = {
        "root_dir": None,
        "ann_dir": None,
        "mask_pointcloud": False,
        "normalize_pointcloud": False,
        "camera_convention": "opengl",
        "scale_convention": "half_max",
        "orientation_repr": "quaternion",
        "orientation_grid_resolution": None,
        "category_str": None,
        "remap_y_axis": None,
        "remap_x_axis": None,
    }

    def __init__(
        self,
        config: Config,
    ) -> None:
        """Initialize the dataset.

        Args:
            config:
                Configuration dictionary of dataset. Provided dictionary will be merged
                with default_dict. See AnnotatedRedwoodDataset.Config for keys.
        """
        config = yoco.load_config(
            config, current_dict=AnnotatedRedwoodDataset.default_config
        )
        self._root_dir = config["root_dir"]
        self._ann_dir = config["ann_dir"]
        self._camera_convention = config["camera_convention"]
        self._mask_pointcloud = config["mask_pointcloud"]
        self._normalize_pointcloud = config["normalize_pointcloud"]
        self._scale_convention = config["scale_convention"]
        self._remap_y_axis = config["remap_y_axis"]
        self._remap_x_axis = config["remap_x_axis"]
        self._orientation_repr = config["orientation_repr"]
        if self._orientation_repr == "discretized":
            self._orientation_grid = so3grid.SO3Grid(
                config["orientation_grid_resolution"]
            )
        self._load_annotations()
        self._camera = Camera(width=640, height=480, fx=525, fy=525, cx=319.5, cy=239.5)

    def _load_annotations(self) -> None:
        """Load annotations into memory."""
        ann_json = os.path.join(self._ann_dir, "annotations.json")
        with open(ann_json, "r") as f:
            anns_dict = json.load(f)
        self._raw_samples = []
        for seq_id, seq_anns in anns_dict.items():
            for pose_ann in seq_anns["pose_anns"]:
                self._raw_samples.append(
                    self._create_raw_sample(seq_id, seq_anns, pose_ann)
                )

    def _create_raw_sample(
        self, seq_id: str, sequence_dict: dict, annotation_dict: dict
    ) -> dict:
        """Create raw sample from information in annotations file."""
        position = torch.tensor(annotation_dict["position"])
        orientation_q = torch.tensor(annotation_dict["orientation"])
        rgb_filename = annotation_dict["rgb_file"]
        depth_filename = annotation_dict["depth_file"]
        mesh_filename = sequence_dict["mesh"]
        mesh_path = os.path.join(self._ann_dir, mesh_filename)
        category_str = sequence_dict["category"]
        color_path = os.path.join(
            self._root_dir, category_str, "rgbd", seq_id, "rgb", rgb_filename
        )
        depth_path = os.path.join(
            self._root_dir, category_str, "rgbd", seq_id, "depth", depth_filename
        )
        extents = torch.tensor(sequence_dict["scale"]) * 2
        return {
            "position": position,
            "orientation_q": orientation_q,
            "extents": extents,
            "color_path": color_path,
            "depth_path": depth_path,
            "mesh_path": mesh_path,
            "category_str": category_str,
        }

    def __len__(self) -> int:
        """Return number of sample in dataset."""
        return len(self._raw_samples)

    def __getitem__(self, idx: int) -> dict:
        """Return a sample of the dataset.

        Args:
            idx: Index of the instance.
        Returns:
            Sample containing the following keys:
                "color"
                "depth"
                "mask"
                "pointset"
                "position"
                "orientation"
                "quaternion"
                "scale"
                "color_path"
                "obj_path"
                "category_id"
                "category_str"
        """
        raw_sample = self._raw_samples[idx]
        color = torch.from_numpy(
            np.asarray(Image.open(raw_sample["color_path"]), dtype=np.float32) / 255
        )
        depth = self._load_depth(raw_sample["depth_path"])
        instance_mask = self._compute_mask(depth, raw_sample)

        pointcloud_mask = instance_mask if self._mask_pointcloud else None
        pointcloud = pointset_utils.depth_to_pointcloud(
            depth,
            self._camera,
            mask=pointcloud_mask,
            convention=self._camera_convention,
        )

        # adjust camera convention for position, orientation and scale
        position = pointset_utils.change_position_camera_convention(
            raw_sample["position"], "opencv", self._camera_convention
        )

        # orientation / scale
        orientation_q, extents = self._change_axis_convention(
            raw_sample["orientation_q"], raw_sample["extents"]
        )
        orientation_q = pointset_utils.change_orientation_camera_convention(
            orientation_q, "opencv", self._camera_convention
        )
        orientation = self._quat_to_orientation_repr(orientation_q)
        scale = self._get_scale(extents)

        # normalize pointcloud & position
        if self._normalize_pointcloud:
            pointcloud, centroid = pointset_utils.normalize_points(pointcloud)
            position = position - centroid

        category_str = raw_sample["category_str"]
        sample = {
            "color": color,
            "depth": depth,
            "pointset": pointcloud,
            "mask": instance_mask,
            "position": position,
            "orientation": orientation,
            "quaternion": orientation_q,
            "scale": scale,
            "color_path": raw_sample["color_path"],
            "obj_path": raw_sample["mesh_path"],
            "category_id": self.category_str_to_id[category_str],
            "category_str": category_str,
        }
        return sample

    def _compute_mask(self, depth: torch.Tensor, raw_sample: dict) -> torch.Tensor:
        mesh = synthetic.Mesh(
            path=raw_sample["mesh_path"],
            scale=1.0,  # do not resize mesh, as it is already at right size
            rel_scale=True,
            center=False,
        )
        mesh.position = raw_sample["position"]
        mesh.orientation = raw_sample["orientation_q"]
        gt_depth = torch.from_numpy(synthetic.draw_depth_geometry(mesh, self._camera))
        mask = gt_depth != 0
        # exclude occluded parts from mask
        mask[(depth != 0) * (depth < gt_depth - 0.01)] = 0
        return mask

    def _load_depth(self, depth_path: str) -> torch.Tensor:
        """Load depth from depth filepath."""
        depth = torch.from_numpy(
            np.asarray(Image.open(depth_path), dtype=np.float32) * 0.001
        )
        return depth

    def _get_scale(self, extents: torch.Tensor) -> float:
        """Return scale from stored sample data and extents."""
        if self._scale_convention == "diagonal":
            return torch.linalg.norm(extents)
        elif self._scale_convention == "max":
            return extents.max()
        elif self._scale_convention == "half_max":
            return 0.5 * extents.max()
        elif self._scale_convention == "full":
            return extents
        else:
            raise ValueError(
                f"Specified scale convention {self._scale_convention} not supported."
            )

    def _change_axis_convention(
        self, orientation_q: torch.Tensor, extents: torch.Tensor
    ) -> tuple:
        """Adjust up-axis for orientation and extents.

        Returns:
            Tuple of position, orienation_q and extents, with specified up-axis.
        """
        if self._remap_y_axis is None and self._remap_x_axis is None:
            return orientation_q, extents
        elif self._remap_y_axis is None or self._remap_x_axis is None:
            raise ValueError("Either both or none of remap_{y,x}_axis have to be None.")

        rotation_o2n = self._get_o2n_object_rotation_matrix()
        remapped_extents = torch.abs(torch.Tensor(rotation_o2n) @ extents)

        # quaternion so far: original -> camera
        # we want a quaternion: new -> camera
        rotation_n2o = rotation_o2n.T

        quaternion_n2o = torch.from_numpy(Rotation.from_matrix(rotation_n2o).as_quat())

        remapped_orientation_q = quaternion_utils.quaternion_multiply(
            orientation_q, quaternion_n2o
        )  # new -> original -> camera

        return remapped_orientation_q, remapped_extents

    def _get_o2n_object_rotation_matrix(self) -> np.ndarray:
        """Compute rotation matrix which rotates original to new object coordinates."""
        rotation_o2n = np.zeros((3, 3))  # original to new object convention
        if self._remap_y_axis == "x":
            rotation_o2n[0, 1] = 1
        elif self._remap_y_axis == "-x":
            rotation_o2n[0, 1] = -1
        elif self._remap_y_axis == "y":
            rotation_o2n[1, 1] = 1
        elif self._remap_y_axis == "-y":
            rotation_o2n[1, 1] = -1
        elif self._remap_y_axis == "z":
            rotation_o2n[2, 1] = 1
        elif self._remap_y_axis == "-z":
            rotation_o2n[2, 1] = -1
        else:
            raise ValueError("Unsupported remap_y_axis {self.remap_y}")

        if self._remap_x_axis == "x":
            rotation_o2n[0, 0] = 1
        elif self._remap_x_axis == "-x":
            rotation_o2n[0, 0] = -1
        elif self._remap_x_axis == "y":
            rotation_o2n[1, 0] = 1
        elif self._remap_x_axis == "-y":
            rotation_o2n[1, 0] = -1
        elif self._remap_x_axis == "z":
            rotation_o2n[2, 0] = 1
        elif self._remap_x_axis == "-z":
            rotation_o2n[2, 0] = -1
        else:
            raise ValueError("Unsupported remap_x_axis {self.remap_y}")

        # infer last column
        rotation_o2n[:, 2] = 1 - np.abs(np.sum(rotation_o2n, 1))  # rows must sum to +-1
        rotation_o2n[:, 2] *= np.linalg.det(rotation_o2n)  # make special orthogonal
        if np.linalg.det(rotation_o2n) != 1.0:  # check if special orthogonal
            raise ValueError("Unsupported combination of remap_{y,x}_axis. det != 1")
        return rotation_o2n

    def _quat_to_orientation_repr(self, quaternion: torch.Tensor) -> torch.Tensor:
        """Convert quaternion to selected orientation representation.

        Args:
            quaternion:
                The quaternion to convert, scalar-last, shape (4,).
        Returns:
            The same orientation as represented by the quaternion in the chosen
            orientation representation.
        """
        if self._orientation_repr == "quaternion":
            return quaternion
        elif self._orientation_repr == "discretized":
            index = self._orientation_grid.quat_to_index(quaternion.numpy())
            return torch.tensor(
                index,
                dtype=torch.long,
            )
        else:
            raise NotImplementedError(
                f"Orientation representation {self._orientation_repr} is not supported."
            )

    def load_mesh(self, object_path: str) -> o3d.geometry.TriangleMesh:
        """Load an object mesh and adjust its object frame convention."""
        mesh = o3d.io.read_triangle_mesh(object_path)
        if self._remap_y_axis is None and self._remap_x_axis is None:
            return mesh
        elif self._remap_y_axis is None or self._remap_x_axis is None:
            raise ValueError("Either both or none of remap_{y,x}_axis have to be None.")

        rotation_o2n = self._get_o2n_object_rotation_matrix()
        mesh.rotate(
            rotation_o2n,
            center=np.array([0.0, 0.0, 0.0])[:, None],
        )
        return mesh


class ObjectError(Exception):
    """Error if something with the mesh is wrong."""

    pass



"""Functions to handle transformations with quaternions.

Inspired by PyTorch3D, but using scalar-last convention and not enforcing scalar > 0.
https://github.com/facebookresearch/pytorch3d
"""
import random
import math

import torch


def quaternion_multiply(
    quaternions_1: torch.Tensor, quaternions_2: torch.Tensor
) -> torch.Tensor:
    """Multiply two quaternions representing rotations.

    Normal broadcasting rules apply.

    Args:
        quaternions_1:
            normalized quaternions of shape (..., 4), scalar-last convention
        quaternions_2:
            normalized quaternions of shape (..., 4), scalar-last convention
    Returns:
        Composition of passed quaternions.
    """
    ax, ay, az, aw = torch.unbind(quaternions_1, -1)
    bx, by, bz, bw = torch.unbind(quaternions_2, -1)
    ox = aw * bx + ax * bw + ay * bz - az * by
    oy = aw * by - ax * bz + ay * bw + az * bx
    oz = aw * bz + ax * by - ay * bx + az * bw
    ow = aw * bw - ax * bx - ay * by - az * bz
    return torch.stack((ox, oy, oz, ow), -1)


def quaternion_apply(quaternions: torch.Tensor, points: torch.Tensor) -> torch.Tensor:
    """Rotate points by quaternions representing rotations.

    Normal broadcasting rules apply.

    Args:
        quaternions:
            normalized quaternions of shape (..., 4), scalar-last convention
        points:
            points of shape (..., 3)
    Returns:
        Points rotated by the rotations representing quaternions.
    """
    points_as_quaternions = points.new_zeros(points.shape[:-1] + (4,))
    points_as_quaternions[..., :-1] = points
    return quaternion_multiply(
        quaternion_multiply(quaternions, points_as_quaternions),
        quaternion_invert(quaternions),
    )[..., :-1]


def quaternion_invert(quaternions: torch.Tensor) -> torch.Tensor:
    """Invert quaternions representing orientations.

    Args:
        quaternions:
            The quaternions to invert, shape (..., 4), scalar-last convention.
    Returns:
        Inverted quaternions, same shape as quaternions.
    """
    return quaternions * quaternions.new_tensor([-1, -1, -1, 1])


def geodesic_distance(q1: torch.Tensor, q2: torch.Tensor) -> torch.Tensor:
    """Compute geodesic distances between quaternions.

    Args:
        q1: First set of quaterions, shape (N,4).
        q2: Second set of quaternions, shape (N,4).
    Returns:
        Mean distance between the quaternions, scalar.
    """
    abs_q1q2 = torch.clip(torch.abs(torch.sum(q1 * q2, dim=1)), 0, 1)
    geodesic_distances = 2 * torch.acos(abs_q1q2)
    return geodesic_distances


def simple_quaternion_loss(q1: torch.Tensor, q2: torch.Tensor) -> torch.Tensor:
    """Compute distance measure between quaternions not involving trig functions.

    From:
        https://math.stackexchange.com/a/90098

    Args:
        q1: First set of quaterions, shape (N,4).
        q2: Second set of quaternions, shape (N,4).
    Returns:
        Mean distance between the quaternions, scalar.
    """
    return torch.mean(1 - torch.sum(q1 * q2, 1) ** 2)


def generate_uniform_quaternion() -> torch.Tensor:
    """Generate a normalized uniform quaternion.

    Following the method from K. Shoemake, Uniform Random Rotations, 1992.

    See: http://planning.cs.uiuc.edu/node198.html

    Returns:
        Uniformly distributed unit quaternion on the estimator's device.
    """
    u1, u2, u3 = random.random(), random.random(), random.random()
    return torch.tensor(
        [
            math.sqrt(1 - u1) * math.sin(2 * math.pi * u2),
            math.sqrt(1 - u1) * math.cos(2 * math.pi * u2),
            math.sqrt(u1) * math.sin(2 * math.pi * u3),
            math.sqrt(u1) * math.cos(2 * math.pi * u3),
        ]
    )



"""Module containing loss functions."""

import torch

from sdfest.initialization import quaternion_utils


def nn_loss(points_from: torch.Tensor, points_to: torch.Tensor) -> torch.Tensor:
    """Compute the distance to the closest neighbor in the other set of points.

    Params:
        points_from:
            The first point set. Shape NxD, with N points of dimension D.
        points_to:
            The second point set. Shape MxD, with M points of dimension D.

    Returns:
        Squared distance from all points in the points_from set to the closest point in
        points to set. Output shape is (N,).
    """
    a = torch.sum(points_from ** 2, dim=1)
    b = torch.mm(points_from, points_to.t())
    c = torch.sum(points_to ** 2, dim=1)

    # compute the distance matrix
    d = -2 * b + a.unsqueeze(1) + c.unsqueeze(0)
    d[d < 0] = 0  # TODO why is it negative sometimes? numerical issues?
    d, _ = d.min(1)
    return d


def pc_loss(
    points: torch.Tensor,
    position: torch.Tensor,
    orientation: torch.Tensor,
    scale: torch.Tensor,
    sdf: torch.Tensor,
) -> torch.Tensor:
    """Compute trilinerly interpolated SDF value at point positions.

    Args:
        points:
            pointcloud in camera frame, shape (M, 4)
        position:
            position of SDF center in camera frame, shape (3,)
        orientation:
            quaternion representing orientation of SDF, shape (4,)
        scale:
            half-width of SDF volume
        sdf:
            volumetric signed distance field, shape (res, res, res),
            assuming same resolution along each axis

    Returns:
        Trilinearly interpolated distance at the passed points 0 if outside of SDF
        volume. Distance is in world coordinates (i.e., after scaling the SDF).
    """
    q = orientation / torch.norm(orientation)  # to get normalization gradients
    obj_points = points - position.unsqueeze(0)

    # Quaternion to rotation matrix
    # Note that we use conjugate here since we want to transform points from
    # world to object coordinates and the quaternion describes rotation of
    # object coordinate system in world coordinates.
    R = obj_points.new_zeros(3, 3)

    R[0, 0] = 1 - 2 * (q[1] * q[1] + q[2] * q[2])
    R[0, 1] = 2 * (q[0] * q[1] + q[2] * q[3])
    R[0, 2] = 2 * (q[0] * q[2] - q[3] * q[1])

    R[1, 0] = 2 * (q[0] * q[1] - q[2] * q[3])
    R[1, 1] = 1 - 2 * (q[0] * q[0] + q[2] * q[2])
    R[1, 2] = 2 * (q[1] * q[2] + q[3] * q[0])

    R[2, 0] = 2 * (q[0] * q[2] + q[3] * q[1])
    R[2, 1] = 2 * (q[1] * q[2] - q[3] * q[0])
    R[2, 2] = 1 - 2 * (q[0] * q[0] + q[1] * q[1])

    obj_points = (R @ obj_points.T).T

    # Transform to canonical coordintes obj_point in [-1,1]^3
    obj_point = obj_points / scale

    # Compute cell and cell position
    res = sdf.shape[0]  # assuming same resolution along each axis
    grid_size = 2.0 / (res - 1)
    c = torch.floor((obj_point + 1.0) * (res - 1) * 0.5)
    mask = torch.logical_or(
        torch.min(c, dim=1)[0] < 0, torch.max(c, dim=1)[0] > res - 2
    )
    c = torch.clip(c, 0, res - 2)  # base cell index of each point
    cell_position = c * grid_size - 1.0  # base cell position of each point
    sdf_indices = c.new_empty((obj_point.shape[0], 8), dtype=torch.long)
    sdf_indices[:, 0] = c[:, 0] * res ** 2 + c[:, 1] * res + c[:, 2]
    sdf_indices[:, 1] = c[:, 0] * res ** 2 + c[:, 1] * res + c[:, 2] + 1
    sdf_indices[:, 2] = c[:, 0] * res ** 2 + (c[:, 1] + 1) * res + c[:, 2]
    sdf_indices[:, 3] = c[:, 0] * res ** 2 + (c[:, 1] + 1) * res + c[:, 2] + 1
    sdf_indices[:, 4] = (c[:, 0] + 1) * res ** 2 + c[:, 1] * res + c[:, 2]
    sdf_indices[:, 5] = (c[:, 0] + 1) * res ** 2 + c[:, 1] * res + c[:, 2] + 1
    sdf_indices[:, 6] = (c[:, 0] + 1) * res ** 2 + (c[:, 1] + 1) * res + c[:, 2]
    sdf_indices[:, 7] = (c[:, 0] + 1) * res ** 2 + (c[:, 1] + 1) * res + c[:, 2] + 1
    sdf_view = sdf.view([-1])
    point_cell_position = (obj_point - cell_position) / grid_size  # [0,1]^3
    sdf_values = torch.take(sdf_view, sdf_indices)

    # trilinear interpolation
    sdf_value = sdf_values.new_empty(obj_points.shape[0])
    # sdf_value = obj_point[:, 0]
    sdf_value = (
        (
            sdf_values[:, 0] * (1 - point_cell_position[:, 0])
            + sdf_values[:, 4] * point_cell_position[:, 0]
        )
        * (1 - point_cell_position[:, 1])
        + (
            sdf_values[:, 2] * (1 - point_cell_position[:, 0])
            + sdf_values[:, 6] * point_cell_position[:, 0]
        )
        * point_cell_position[:, 1]
    ) * (1 - point_cell_position[:, 2]) + (
        (
            sdf_values[:, 1] * (1 - point_cell_position[:, 0])
            + sdf_values[:, 5] * point_cell_position[:, 0]
        )
        * (1 - point_cell_position[:, 1])
        + (
            sdf_values[:, 3] * (1 - point_cell_position[:, 0])
            + sdf_values[:, 7] * point_cell_position[:, 0]
        )
        * point_cell_position[:, 1]
    ) * point_cell_position[
        :, 2
    ]
    sdf_value[mask] = 0
    return sdf_value * scale


def point_constraint_loss(
    orientation_q: torch.Tensor, source: torch.Tensor, target: torch.Tensor
) -> torch.Tensor:
    """Compute Euclidean distance between rotated source point and target point.

    Args:
        orientation_q:
            Orientation of object in world / camera frame as quaternion.
            Scalar-last convention. Shape (4,).
        source: Point in object frame, which will be transformed. (3,).
        target: Point in rotated oject frame. Shape (3,).
    Returns:
        Euclidean norm between R(orientation_q) @ source - target. Scalar.
    """
    rotated_source = quaternion_utils.quaternion_apply(orientation_q, source)
    return torch.linalg.norm(rotated_source - target)



"""Utility functions to handle voxel-based signed distance fields."""
from typing import Optional

import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from scipy.spatial.transform import Rotation


def sdf_to_pointcloud(
    sdf: np.array,
    position: np.array,
    orientation: np.array,
    scale: float,
    max_points: Optional[int] = None,
    threshold: float = 0,
):
    """Convert SDF to pointcloud.

    Puts a point onto each cell vertex with a value < threshold.

    Args:
        sdf: The values of the voxelized SDF. Shape (D, D, D).
        position: The position of the SDF center. Shape (3,).
        orientation:
            The orientation of the SDF as a normalized quaternion.
            This is the quaternion that will be applied to each point.
            Scalar-last convention, shape (4,).
        scale: The half-length of the SDF.
        max_points: Maximum number of points in the pointcloud.
        threshold: The threshold below which a voxel will be included in the pointcloud.

    Returns:
        The pointcloud as a Nx3 array.
    """
    grid_size = 2.0 / (sdf.shape[0] - 1.0)
    indices = np.argwhere(sdf <= threshold)
    points = (indices * grid_size - 1.0) * scale
    rot_matrix = Rotation.from_quat(orientation).as_matrix()
    points = (rot_matrix @ points.T).T
    points += position
    if max_points is not None and max_points < points.shape[0]:
        points = points[np.random.choice(points.shape[0], 2, replace=False), :]
    return points


if __name__ == "__main__":
    sdf = np.random.random((10, 10, 10))
    pos = np.array([1.0, 0.5, 0.0])
    orientation = np.array([0.0, 0.0, 0.0, 1.0])
    orientation /= np.linalg.norm(orientation)
    points = sdf_to_pointcloud(sdf, pos, orientation, 0.15, 0.5)
    fig = plt.figure()
    ax = fig.add_subplot(111, projection=Axes3D.name)
    ax.scatter(points[:, 0], points[:, 1], points[:, 2])
    plt.show()



"""Modular SDF pose and shape estimation in depth images."""
import copy
import math
import os
import pickle
import random
import time
from typing import Optional, Tuple

import ffmpeg
import matplotlib.pyplot as plt
import numpy as np
import open3d as o3d
from skimage.measure import marching_cubes
from sdfest.vae import sdf_utils
from sdfest.vae.sdf_vae import SDFVAE
from sdfest.initialization.sdf_pose_network import SDFPoseNet, SDFPoseHead
from sdfest.initialization.pointnet import VanillaPointNet
from sdfest.initialization import pointset_utils, quaternion_utils
from sdfest.differentiable_renderer import Camera, render_depth_gpu
import torch

from sdfest.estimation import synthetic, losses
from sdfest.utils import load_model_weights


INIT_MODULE_DICT = {c.__name__: c for c in [SDFPoseHead, VanillaPointNet]}


class NoDepthError(ValueError):
    """Raised when there is no depth data left after preprocessing."""
    pass


class SDFPipeline:
    """SDF pose and shape estimation pipeline."""

    def __init__(self, config: dict) -> None:
        """Load and initialize the pipeline.

        Args:
            config: Configuration dictionary.
        """
        # improve runtime for conv3d backward
        # https://github.com/pytorch/pytorch/issues/32370
        torch.backends.cudnn.enabled = False

        self._parse_config(config)

        self.init_network = SDFPoseNet(
            INIT_MODULE_DICT[self.init_config["backbone_type"]](
                **self.init_config["backbone"]
            ),
            INIT_MODULE_DICT[self.init_config["head_type"]](
                shape_dimension=self.vae_config["latent_size"],
                **self.init_config["head"],
            ),
        ).to(self.device)
        load_model_weights(
            self.init_config["model"],
            self.init_network,
            self.device,
            self.init_config.get("model_url"),
        )
        self.init_network.eval()

        self.resolution = 64
        self.vae = SDFVAE(
            sdf_size=64,
            latent_size=self.vae_config["latent_size"],
            encoder_dict=self.vae_config["encoder"],
            decoder_dict=self.vae_config["decoder"],
            device=self.device,
        ).to(self.device)
        load_model_weights(
            self.vae_config["model"],
            self.vae,
            self.device,
            self.vae_config.get("model_url"),
        )
        self.vae.eval()

        self.cam = Camera(**self.camera_config)
        self.render = lambda sdf, pos, quat, i_s: render_depth_gpu(
            sdf, pos, quat, i_s, None, None, None, config["threshold"], self.cam
        )
        self.config = config

        self.log_data = []

    def _parse_config(self, config: dict) -> None:
        """Parse config dict.

        This function makes sure that all required keys are available.
        """
        self.device = config["device"]
        self.init_config = config["init"]
        self.vae_config = config["vae"] if "vae" in config else self.init_config["vae"]
        self.camera_config = config["camera"]
        self.result_selection_strategy = config.get(
            "result_selection_strategy", "last_iteration"
        )  # last_iteration | best_inlier_ratio
        self._relative_inlier_threshold = config.get(
            "relative_inlier_threshold", 0.03
        )  # relative depth error threshold for pixel to be considered inlier
        if "far_field" in config:
            self._far_field = config["far_field"] if "far_field" in config else None

        self.config = config

    @staticmethod
    def _compute_gradients(loss: torch.Tensor) -> None:
        loss.backward()

    def _compute_view_losses(
        self,
        depth_input: torch.Tensor,
        depth_estimate: torch.Tensor,
        position: torch.Tensor,
        orientation: torch.Tensor,
        scale: torch.Tensor,
        sdf: torch.Tensor,
    ) -> Tuple[torch.Tensor]:
        # depth l1
        overlap_mask = (depth_input > 0) & (depth_estimate > 0)
        depth_error = torch.abs(depth_estimate - depth_input)
        # max_depth_error = 0.05
        # depth_outlier_mask = depth_error > max_depth_error
        # depth_mask = overlap_mask & ~depth_outlier_mask
        # depth_error[~overlap_mask] = 0
        loss_depth = torch.mean(depth_error[overlap_mask])

        # pointcloud l1
        pointcloud_obs = pointset_utils.depth_to_pointcloud(
            depth_input, self.cam, normalize=False
        )
        pointcloud_error = losses.pc_loss(
            pointcloud_obs,
            position,
            orientation,
            scale,
            sdf,
        )
        loss_pc = torch.mean(torch.abs(pointcloud_error))

        # nearest neighbor l1
        # pointcloud_outliers = pointset_utils.depth_to_pointcloud(
        #     depth_estimate, self.cam, normalize=False, mask=depth_outlier_mask
        # )

        loss_nn = 0

        # if pointcloud_outliers.shape[0] != 0:
        # pass
        # loss_nn += 0
        # # TODO different gradients for point cloud (not derived by renderer)
        # outlier_nn_d = losses.nn_loss(pointcloud_outliers, pointcloud_obs)
        # # only use positive, because sqrt is not differentiable at 0
        # outlier_nn_d = outlier_nn_d[outlier_nn_d > 0]
        # loss_nn = loss_nn + torch.mean(torch.sqrt(outlier_nn_d))

        return loss_depth, loss_pc, loss_nn

    def _compute_point_constraint_loss(self, orientation: torch.Tensor) -> torch.Tensor:
        """Compute loss for point constraint if specified."""
        if self._point_constraint is not None:
            loss_point_constraint = losses.point_constraint_loss(
                orientation_q=orientation[0],
                source=self._point_constraint[0],
                target=self._point_constraint[1],
            )
            weight = self._point_constraint[2]
            return weight * loss_point_constraint
        else:
            return orientation.new_tensor(0.0)

    def _compute_inlier_ratio(
        self,
        depth_input: torch.Tensor,
        depth_estimate: torch.Tensor,
    ) -> None:
        """Compute ratio of pixels with small relative depth error."""
        rel_depth_error = torch.abs(depth_input - depth_estimate) / depth_input
        inlier_mask = rel_depth_error < self._relative_inlier_threshold
        inliers = torch.count_nonzero(inlier_mask)
        valid_depth_pixels = torch.count_nonzero(depth_input)
        inlier_ratio = inliers / valid_depth_pixels
        return inlier_ratio

    def _update_best_estimate(
        self,
        depth_input: torch.Tensor,
        depth_estimate: torch.Tensor,
        position: torch.Tensor,
        orientation: torch.Tensor,
        scale: torch.Tensor,
        latent_shape: torch.Tensor,
    ) -> None:
        """Update the best current estimate by keeping track of inlier ratio.

        Returns:
            Inlier ratio of this configuration.
        """
        inlier_ratio = self._compute_inlier_ratio(depth_input, depth_estimate)
        if self._best_inlier_ratio is None or inlier_ratio > self._best_inlier_ratio:
            self._best_inlier_ratio = inlier_ratio
            self._best_position = position
            self._best_orientation = orientation
            self._best_scale = scale
            self._best_latent_shape = latent_shape
        return inlier_ratio

    def __call__(
        self,
        depth_images: torch.Tensor,
        masks: torch.Tensor,
        color_images: torch.Tensor,
        visualize: bool = False,
        camera_positions: Optional[torch.Tensor] = None,
        camera_orientations: Optional[torch.Tensor] = None,
        log_path: Optional[str] = None,
        shape_optimization: bool = True,
        animation_path: Optional[str] = None,
        point_constraint: Optional[Tuple[torch.Tensor]] = None,
        prior_orientation_distribution: Optional[torch.Tensor] = None,
        training_orientation_distribution: Optional[torch.Tensor] = None,
    ) -> tuple:
        """Infer pose, size and latent representation from depth and mask.

        If multiple images are passed the cameras are assumed to be fixed.
        All tensors should be on the same device as the pipeline.

        Batch dimension N must be provided either for all or none of the arguments.

        Args:
            depth_images:
                The depth map containing the distance along the camera's z-axis.
                Does not have to be masked, necessary preprocessing is done by pipeline.
                Will be masked and preprocessed in-place (pass copy if full depth is
                used afterwards).
                Shape (N, H, W) or (H, W) for a single depth image.
            masks:
                binary mask of the object to estimate, same shape as depth_images
            color_images:
                the color image (currently only used in visualization),
                shape (N, H, W, 3) or (H, W, 3), RGB, float, 0-1.
            visualize: Whether to visualize the intermediate steps and final result.
            camera_positions:
                position of camera in world coordinates for each image,
                if None, (0,0,0) will be assumed for all images,
                shape (N, 3) or (3,)
            camera_orientations:
                orientation of camera in world-frame as normalized quaternion,
                quaternion is in scalar-last convention,
                note, that this is the quaternion that transforms a point from camera
                to world-frame
                if None, (0,0,0,1) will be assumed for all images,
                shape (N, 4) or (4,)
            log_path:
                file path to write timestamps and intermediate steps to,
                no logging is performed if None
            shape_optimization:
                enable or disable shape optimization during iterative optimization
            animation_path:
                file path to write rendering and error visualizations to
            point_constraint:
                tuple of source point and rotated target point and weight
                a loss will be added that penalizes
                    weight * || rotation @ source - target ||_2
            prior_orientation_distribution:
                Prior distribution of orientations used for initialization.
                If None, distribution of initialization network will not be modified.
                Only supported for initialization network with discretized orientation
                representation.
                Output distribution of initialization network will be adjusted by
                multiplying with
                prior_orientation_distribution / training_orientation_distribution
                and renormalizing.
                Tensor of shape (N,C,) or (C,) for single image.
                C being the number of grid cells in the SO3Grid used by the
                initialization network.
            training_orientation_distribution:
                Distribution of orientations used for training initialization network.
                If None, equal probability for each cell will be assumed.
                Note this is only approximately the same as a uniform distribution.
                Only used if prior_orientation_distribution is provided.
                Tensor of shape (C,). C being the number of grid cells in the SO3Grid
                used by the initialization network. N not supported, since same
                training network (and hence distribution) is used independent of view.

        Returns:
            - 3D pose of SDF center in world frame, shape (1,3,)
            - Orientation as normalized quaternion, scalar-last convention, shape (1,4,)
            - Size of SDF as length of half-width, shape (1,)
            - Latent shape representation of the object, shape (1,latent_size,).
        """
        # initialize optimization
        self._best_inlier_ratio = None
        self._point_constraint = point_constraint

        if animation_path is not None:
            self._create_animation_folders(animation_path)

        start_time = time.time()  # for logging

        # Add batch dimension if necessary
        if depth_images.dim() == 2:
            depth_images = depth_images.unsqueeze(0)
            masks = masks.unsqueeze(0)
            color_images = color_images.unsqueeze(0)
            if camera_positions is not None:
                camera_positions = camera_positions.unsqueeze(0)
            if camera_orientations is not None:
                camera_orientations = camera_orientations.unsqueeze(0)
            if prior_orientation_distribution is not None:
                prior_orientation_distribution = (
                    prior_orientation_distribution.unsqueeze(0)
                )

        # TODO assert all tensors have expected dimension

        if animation_path is not None:
            self._save_inputs(animation_path, depth_images, color_images, masks)

        n_imgs = depth_images.shape[0]

        if camera_positions is None:
            camera_positions = torch.zeros(n_imgs, 3, device=self.device)
        if camera_orientations is None:
            camera_orientations = torch.zeros(n_imgs, 4, device=self.device)
            camera_orientations[:, 3] = 1.0

        with torch.no_grad():
            self._preprocess_depth(depth_images, masks)

        # store pointcloud without reconstruction
        if log_path is not None:
            torch.cuda.synchronize()
            self._log_data(
                {
                    "timestamp": time.time() - start_time,
                    "depth_images": depth_images,
                    "color_images": color_images,
                    "masks": masks,
                    "color_images": color_images,
                    "camera_positions": camera_positions,
                    "camera_orientations": camera_orientations,
                },
            )

        # Initialization
        with torch.no_grad():
            latent_shape, position, scale, orientation = self._nn_init(
                depth_images,
                camera_positions,
                camera_orientations,
                prior_orientation_distribution,
                training_orientation_distribution,
            )

        if log_path is not None:
            torch.cuda.synchronize()
            self._log_data(
                {
                    "timestamp": time.time() - start_time,
                    "camera_positions": camera_positions,
                    "camera_orientations": camera_orientations,
                    "latent_shape": latent_shape,
                    "position": position,
                    "scale_inv": 1 / scale,
                    "orientation": orientation,
                },
            )

        if animation_path is not None:
            self._save_preprocessed_inputs(animation_path, depth_images)

        # Iterative optimization
        self._current_iteration = 1

        position.requires_grad_()
        scale.requires_grad_()
        orientation.requires_grad_()
        latent_shape.requires_grad_()

        if visualize:
            fig_vis, axes = plt.subplots(
                2, 3, sharex=True, sharey=True, figsize=(12, 8)
            )
            fig_loss, (loss_ax, inlier_ax) = plt.subplots(1, 2)
            vmin, vmax = None, None

        depth_losses = []
        pointcloud_losses = []
        nn_losses = []
        point_constraint_losses = []
        inlier_ratios = []
        total_losses = []

        opt_vars = [
            {"params": position, "lr": 1e-3},
            {"params": orientation, "lr": 1e-2},
            {"params": scale, "lr": 1e-3},
            {"params": latent_shape, "lr": 1e-2},
        ]
        optimizer = torch.optim.Adam(opt_vars)

        while self._current_iteration <= self.config["max_iterations"]:
            optimizer.zero_grad()

            norm_orientation = orientation / torch.sqrt(torch.sum(orientation**2))

            with torch.set_grad_enabled(shape_optimization):
                sdf = self.vae.decode(latent_shape)

            loss_depth = torch.tensor(0.0, device=self.device, requires_grad=True)
            loss_pc = torch.tensor(0.0, device=self.device, requires_grad=True)
            loss_nn = torch.tensor(0.0, device=self.device, requires_grad=True)

            for depth_image, camera_position, camera_orientation in zip(
                depth_images, camera_positions, camera_orientations
            ):
                # transform object to camera frame
                q_w2c = quaternion_utils.quaternion_invert(camera_orientation)
                position_c = quaternion_utils.quaternion_apply(
                    q_w2c, position - camera_position
                )
                orientation_c = quaternion_utils.quaternion_multiply(
                    q_w2c, norm_orientation
                )

                depth_estimate = self.render(
                    sdf[0, 0], position_c[0], orientation_c[0], 1 / scale[0]
                )

                view_loss_depth, view_loss_pc, view_loss_nn = self._compute_view_losses(
                    depth_image,
                    depth_estimate,
                    position_c[0],
                    orientation_c[0],
                    scale[0],
                    sdf[0, 0],
                )
                loss_depth = loss_depth + view_loss_depth
                loss_pc = loss_pc + view_loss_pc
                loss_nn = loss_nn + view_loss_nn

            loss_point_constraint = self._compute_point_constraint_loss(orientation)
            loss = (
                self.config["depth_weight"] * loss_depth
                + self.config["pc_weight"] * loss_pc
                + self.config["nn_weight"] * loss_nn
                + loss_point_constraint
            )

            self._compute_gradients(loss)

            optimizer.step()
            optimizer.zero_grad()

            with torch.no_grad():
                orientation /= torch.sqrt(torch.sum(orientation**2))
                inlier_ratio = self._update_best_estimate(
                    depth_image,
                    depth_estimate,
                    position,
                    orientation,
                    scale,
                    latent_shape,
                )

            if visualize:
                depth_losses.append(loss_depth.item())
                pointcloud_losses.append(loss_pc.item())
                nn_losses.append(loss_nn.item())
                point_constraint_losses.append(loss_point_constraint.item())
                inlier_ratios.append(inlier_ratio.item())
                total_losses.append(loss.item())

            if log_path is not None:
                torch.cuda.synchronize()
                self._log_data(
                    {
                        "timestamp": time.time() - start_time,
                        "latent_shape": latent_shape,
                        "position": position,
                        "scale_inv": 1 / scale,
                        "orientation": orientation,
                    },
                )

            with torch.no_grad():
                if animation_path is not None:
                    self._save_current_state(
                        depth_images,
                        animation_path,
                        camera_positions,
                        camera_orientations,
                        position,
                        orientation,
                        1 / scale,
                        sdf,
                    )

                if visualize and (
                    self._current_iteration % 10 == 1
                    or self._current_iteration == self.config["max_iterations"]
                ):
                    q_w2c = quaternion_utils.quaternion_invert(camera_orientations[0])
                    position_c = quaternion_utils.quaternion_apply(
                        q_w2c, position - camera_positions[0]
                    )
                    orientation_c = quaternion_utils.quaternion_multiply(
                        q_w2c, orientation
                    )

                    current_depth = self.render(
                        sdf[0, 0], position_c, orientation_c, 1 / scale
                    )

                    depth_image = depth_images[0]
                    color_image = color_images[0]

                    if self._current_iteration == 1:
                        vmin = depth_image[depth_image != 0].min() * 0.9
                        vmax = depth_image[depth_image != 0].max()
                        # show input image
                        axes[0, 0].clear()
                        axes[0, 0].imshow(depth_image.cpu(), vmin=vmin, vmax=vmax)
                        axes[0, 1].imshow(color_image.cpu())

                        # show initial estimate
                        axes[1, 0].clear()
                        axes[1, 0].imshow(
                            current_depth.detach().cpu(), vmin=vmin, vmax=vmax
                        )
                        axes[1, 0].set_title(f"loss {loss.item()}")

                    # update iterative estimate
                    # axes[0, 2].clear()
                    # axes[0, 2].imshow(rendered_error.detach().cpu())
                    # axes[0, 2].set_title("depth_loss")
                    # axes[1, 2].clear()
                    # axes[1, 2].imshow(error_pc.detach().cpu())
                    # axes[1, 2].set_title("pointcloud_loss")

                    loss_ax.clear()
                    loss_ax.plot(depth_losses, label="Depth")
                    loss_ax.plot(pointcloud_losses, label="Pointcloud")
                    loss_ax.plot(nn_losses, label="Nearest Neighbor")
                    if self._point_constraint is not None:
                        loss_ax.plot(point_constraint_losses, label="Point constraint")
                    loss_ax.plot(total_losses, label="Total")
                    loss_ax.set_yscale("log")
                    loss_ax.legend()

                    inlier_ax.clear()
                    inlier_ax.plot(inlier_ratios, label="Inlier Ratio")
                    inlier_ax.legend()

                    axes[1, 1].clear()
                    axes[1, 1].imshow(
                        current_depth.detach().cpu(), vmin=vmin, vmax=vmax
                    )
                    axes[1, 1].set_title(f"loss {loss.item()}")
                    fig_loss.canvas.draw()
                    fig_vis.canvas.draw()
                    plt.pause(0.1)

            self._current_iteration += 1

        if visualize:
            plt.show()
            plt.close(fig_loss)
            plt.close(fig_vis)

        if log_path is not None:
            self._write_log_data(log_path)

        if animation_path is not None:
            self._create_animations(animation_path)

        if self.result_selection_strategy == "last_iteration":
            return position, orientation, scale, latent_shape
        elif self.result_selection_strategy == "best_inlier_ratio":
            return (
                self._best_position,
                self._best_orientation,
                self._best_scale,
                self._best_latent_shape,
            )
        else:
            raise ValueError(
                f"Result selection strategy {self.result_selection_strategy} is not"
                "supported."
            )

    def _log_data(self, data: dict) -> None:
        """Add dictionary with associated timestamp to log data list."""
        new_log_data = copy.deepcopy(data)
        self.log_data.append(new_log_data)

    def _write_log_data(self, file_path: str) -> None:
        """Write current list of log data to file."""
        with open(file_path, "wb") as f:
            pickle.dump({"config": self.config, "log": self.log_data}, f)
        self.log_data = []  # reset log

    def generate_depth(
        self,
        position: torch.Tensor,
        orientation: torch.Tensor,
        scale: torch.Tensor,
        latent: torch.Tensor,
    ) -> torch.Tensor:
        """Generate depth image representing positioned object."""
        sdf = self.vae.decode(latent)
        depth = self.render(sdf[0, 0], position, orientation, 1 / scale)
        return depth

    def generate_mesh(
        self, latent: torch.tensor, scale: torch.tensor, complete_mesh: bool = False
    ) -> synthetic.Mesh:
        """Generate mesh without pose.

        Currently only supports batch size 1.

        Args:
            latent: Latent shape descriptor, shape (1,L).
            scale:
                Relative scale of the signed distance field, (i.e., half-width),
                shape (1,).
            complete_mesh:
                If True, the SDF will be padded with positive values prior to converting
                it to a mesh. This ensures a watertight mesh is created.

        Returns:
            Generate mesh by decoding latent shape descriptor and scaling it.
        """
        with torch.no_grad():
            sdf = self.vae.decode(latent)
            if complete_mesh:
                inc = 2
                sdf = torch.nn.functional.pad(sdf, (1, 1, 1, 1, 1, 1), value=1.0)
            else:
                inc = 0
            try:
                sdf = sdf.cpu().numpy()
                s = 2.0 / (self.resolution - 1)
                vertices, faces, _, _ = marching_cubes(
                    sdf[0, 0],
                    spacing=(
                        s,
                        s,
                        s,
                    ),
                    level=self.config["iso_threshold"],
                )

                c = s * (self.resolution + inc - 1) / 2.0  # move origin to center
                vertices -= np.array([[c, c, c]])

                mesh = o3d.geometry.TriangleMesh(
                    vertices=o3d.utility.Vector3dVector(vertices),
                    triangles=o3d.utility.Vector3iVector(faces),
                )
            except KeyError:
                return None
            return synthetic.Mesh(mesh=mesh, scale=scale.item(), rel_scale=True)

    def _preprocess_depth(
        self, depth_images: torch.Tensor, masks: torch.Tensor
    ) -> None:
        """Preprocesses depth image based on segmentation mask.

        Args:
            depth_images:
                the depth images to preprocess, will be modified in place,
                shape (N, H, W)
            masks: the masks used for preprocessing, same shape as depth_images
        """
        # shrink mask
        # masks = (
        #     -torch.nn.functional.max_pool2d(
        #         -masks.double(), kernel_size=9, stride=1, padding=4
        #     )
        # ).bool()

        depth_images[~masks] = 0  # set outside of depth to 0

        # remove data far away (should be based on what distances ocurred in training)
        if self._far_field is not None:
            depth_images[depth_images > self._far_field] = 0

        # only consider available depth values for outlier detection
        # masks = torch.logical_and(masks, depth_images != 0)

        # depth_images =

        # remove outliers based on median
        # plt.imshow(depth_images[0].cpu().numpy())
        # plt.show()
        # for mask, depth_image in zip(masks, depth_images):
        #     median = torch.median(depth_image[mask])
        #     errors = torch.abs(depth_image[mask] - median)

        #     bins = 100
        #     hist = torch.histc(errors, bins=bins)
        #     print(hist)
        #     zero_indices = torch.nonzero(hist == 0)
        #     if len(zero_indices):
        #         threshold = zero_indices[0] / bins * errors.max()
        #         print(threshold)
        #         depth_image[torch.abs(depth_image - median) > threshold] = 0
        # plt.imshow(depth_images[0].cpu().numpy())
        # plt.show()

    def _nn_init(
        self,
        depth_images: torch.Tensor,
        camera_positions: torch.Tensor,
        camera_orientations: torch.Tensor,
        prior_orientation_distribution: Optional[torch.Tensor] = None,
        training_orientation_distribution: Optional[torch.Tensor] = None,
    ) -> Tuple:
        """Estimate shape, pose, scale and orientation using initialization network.

        Args:
            depth_images: the preprocessed depth images, shape (N, H, W)
            camera_positions:
                position of camera in world coordinates for each image, shape (N, 3)
            camera_orientations:
                orientation of camera in world-frame as normalized quaternion,
                quaternion is in scalar-last convention, shape (N, 4)
            prior_orientation_distribution:
                Prior distribution of orientations used for initialization.
                If None, distribution of initialization network will not be modified.
                Only supported for initialization network with discretized orientation
                representation.
                Output distribution of initialization network will be adjusted by
                multiplying with
                prior_orientation_distribution / training_orientation_distribution
                and renormalizing.
                Tensor of shape (N,C,). C being the number of grid cells in the SO3Grid
                used by the initialization network.
            training_orientation_distribution:
                Distribution of orientations used for training initialization network.
                If None, equal probability for each cell will be assumed.
                Note this is only approximately the same as a uniform distribution.
                Only used if prior_orientation_distribution is provided.
                Tensor of shape (C,). C being the number of grid cells in the SO3Grid
                used by the initialization network.

        Returns:
            Tuple comprised of:
            - Latent shape representation of the object, shape (1, latent_size)
            - 3D pose of SDF center in camera frame, shape (1, 3)
            - Size of SDF as length of half-width, (1,)
            - Orientation of SDF as normalized quaternion (1,4)
        """
        if (
            prior_orientation_distribution is not None
            and self.init_config["head"]["orientation_repr"] != "discretized"
        ):
            raise ValueError(
                "prior_orientation_distribution only supported for discretized "
                "orientation representation."
            )

        best = 0
        best_result = None
        for i, (depth_image, camera_orientation, camera_position) in enumerate(
            zip(depth_images, camera_orientations, camera_positions)
        ):
            centroid = None
            if self.init_config["backbone_type"] == "VanillaPointNet":
                inp = pointset_utils.depth_to_pointcloud(
                    depth_image, self.cam, normalize=False
                )
                if len(inp) == 0:
                    raise NoDepthError
                if self.init_config["normalize_pose"]:
                    inp, centroid = pointset_utils.normalize_points(inp)
            else:
                inp = depth_image

            inp = inp.unsqueeze(0)
            latent_shape, position, scale, orientation_repr = self.init_network(inp)

            if self.config["mean_shape"]:
                latent_shape = latent_shape.new_zeros(latent_shape.shape)

            if centroid is not None:
                position += centroid

            if self.init_config["head"]["orientation_repr"] == "discretized":
                posterior_orientation_dist = torch.softmax(orientation_repr, -1)

                if prior_orientation_distribution is not None:
                    posterior_orientation_dist = self._adjust_categorical_posterior(
                        posterior=posterior_orientation_dist,
                        prior=prior_orientation_distribution[i],
                        train_prior=training_orientation_distribution,
                    )

                orientation_camera = torch.tensor(
                    self.init_network._head._grid.index_to_quat(
                        posterior_orientation_dist.argmax().item()
                    ),
                    dtype=torch.float,
                    device=self.device,
                ).unsqueeze(0)
            elif self.init_config["head"]["orientation_repr"] == "quaternion":
                orientation_camera = orientation_repr
            else:
                raise NotImplementedError("Orientation representation is not supported")

            # output are in camera frame, transform to world frame
            position_world = (
                quaternion_utils.quaternion_apply(camera_orientation, position)
                + camera_position
            )
            orientation_world = quaternion_utils.quaternion_multiply(
                camera_orientation, orientation_camera
            )

            if self.config["init_view"] == "first":
                return latent_shape, position_world, scale, orientation_world
            elif self.config["init_view"] == "best":
                if self.init_config["head"]["orientation_repr"] != "discretized":
                    raise NotImplementedError(
                        '"best" init strategy only supported with discretized '
                        "orientation representation"
                    )
                maximum = posterior_orientation_dist.max()
                if maximum > best:
                    best = maximum
                    best_result = latent_shape, position_world, scale, orientation_world
            else:
                raise NotImplementedError(
                    'Only "first" and "best" strategies are currently supported'
                )

        return best_result

    def _generate_uniform_quaternion(self) -> torch.tensor:
        """Generate a uniform quaternion.

        Following the method from K. Shoemake, Uniform Random Rotations, 1992.

        See: http://planning.cs.uiuc.edu/node198.html

        Returns:
            Uniformly distributed unit quaternion on the estimator's device.
        """
        u1, u2, u3 = random.random(), random.random(), random.random()
        return (
            torch.tensor(
                [
                    math.sqrt(1 - u1) * math.sin(2 * math.pi * u2),
                    math.sqrt(1 - u1) * math.cos(2 * math.pi * u2),
                    math.sqrt(u1) * math.sin(2 * math.pi * u3),
                    math.sqrt(u1) * math.cos(2 * math.pi * u3),
                ]
            )
            .unsqueeze(0)
            .to(self.device)
        )

    def _create_animation_folders(self, animation_path: str) -> None:
        """Create subfolders to store animation frames."""
        os.makedirs(animation_path)
        depth_path = os.path.join(animation_path, "depth")
        os.makedirs(depth_path)
        error_path = os.path.join(animation_path, "depth_error")
        os.makedirs(error_path)
        sdf_path = os.path.join(animation_path, "sdf")
        os.makedirs(sdf_path)

    def _save_inputs(
        self,
        animation_path: str,
        color_images: torch.Tensor,
        depth_images: torch.Tensor,
        instance_masks: torch.Tensor,
    ) -> None:
        color_path = os.path.join(animation_path, "color_input.png")
        fig, ax = plt.subplots()
        ax.imshow(color_images[0].cpu().numpy())
        fig.savefig(color_path)
        plt.close(fig)
        depth_path = os.path.join(animation_path, "depth_input.png")
        fig, ax = plt.subplots()
        ax.imshow(depth_images[0].cpu().numpy())
        fig.savefig(depth_path)
        plt.close(fig)
        mask_path = os.path.join(animation_path, "mask.png")
        fig, ax = plt.subplots()
        ax.imshow(instance_masks[0].cpu().numpy())
        fig.savefig(mask_path)
        plt.close(fig)

    def _save_preprocessed_inputs(
        self,
        animation_path: str,
        depth_images: torch.Tensor,
    ) -> None:
        depth_path = os.path.join(animation_path, "preprocessed_depth_input.png")
        fig, ax = plt.subplots()
        ax.imshow(depth_images[0].cpu().numpy())
        fig.savefig(depth_path)
        plt.close(fig)

    def _save_current_state(
        self,
        depth_images: torch.Tensor,
        animation_path: str,
        camera_positions: torch.Tensor,
        camera_orientations: torch.Tensor,
        position: torch.Tensor,
        orientation: torch.Tensor,
        scale_inv: torch.Tensor,
        sdf: torch.Tensor,
    ) -> None:
        q_w2c = quaternion_utils.quaternion_invert(camera_orientations[0])
        position_c = quaternion_utils.quaternion_apply(
            q_w2c, position - camera_positions[0]
        )
        orientation_c = quaternion_utils.quaternion_multiply(q_w2c, orientation)
        current_depth = self.render(sdf[0, 0], position_c, orientation_c, scale_inv)
        depth_path = os.path.join(
            animation_path, "depth", f"{self._current_iteration:06}.png"
        )
        fig, ax = plt.subplots()
        ax.imshow(current_depth.cpu().numpy(), interpolation="none")
        fig.savefig(depth_path)
        plt.close(fig)

        error_image = torch.abs(current_depth - depth_images[0])
        error_image[depth_images[0] == 0] = 0
        error_image[current_depth == 0] = 0
        error_path = os.path.join(
            animation_path, "depth_error", f"{self._current_iteration:06}.png"
        )
        fig, ax = plt.subplots()
        ax.imshow(error_image.cpu().numpy(), interpolation="none")
        fig.savefig(error_path)
        plt.close(fig)

        unscaled_threshold = self.config["threshold"] * scale_inv.item()
        mesh = sdf_utils.mesh_from_sdf(
            sdf[0, 0].cpu().numpy(),
            unscaled_threshold,
            complete_mesh=True,
        )
        sdf_path = os.path.join(
            animation_path, "sdf", f"{self._current_iteration:06}.png"
        )

        # map y -> z; z -> y
        transform = np.eye(4)
        transform[0:3, 0:3] = np.array([[-1, 0, 0], [0, 0, 1], [0, 1, 0]])
        fig, ax = plt.subplots()
        sdf_utils.plot_mesh(mesh, transform=transform, plot_object=ax)
        fig.savefig(sdf_path)
        plt.close(fig)

    def _create_animations(self, animation_path: str) -> None:
        names = ["sdf", "depth", "depth_error"]
        for name in names:
            frame_folder = os.path.join(animation_path, name)
            video_name = os.path.join(animation_path, f"{name}.mp4")
            ffmpeg.input(
                os.path.join(frame_folder, "*.png"), pattern_type="glob", framerate=30
            ).output(video_name).run()

    @staticmethod
    def _adjust_categorical_posterior(
        posterior: torch.Tensor, prior: torch.Tensor, train_prior: torch.Tensor
    ) -> torch.Tensor:
        """Adjust categorical posterior distribution.

        Posterior is calculated with a train_prior

        Args:
            posterior:
                Posterior distribution computed assuming train_prior.
                Shape (..., K). K being number of categories.
            prior:
                The desired new prior distribution.
                Same shape as posterior.
            train_prior:
                The prior distribution used to compute the posterior.
                If None, equal probability for each category will be assumed.
                Same shape as posterior.

        Returns:
            The categorical posterior, adjusted such that prior is prior, instead of
            train_prior.
            Same shape as posterior.
        """
        adjusted_posterior = posterior.clone()
        # adjust if prior different from training
        adjusted_posterior *= prior
        if train_prior is not None:
            adjusted_posterior /= train_prior
        adjusted_posterior = torch.nn.functional.normalize(
            adjusted_posterior, p=1, dim=-1
        )
        return adjusted_posterior



"""Utility functions to handle pointsets."""
import torch
from typing import Optional

import matplotlib.pyplot as plt
import numpy as np

from sdfest.differentiable_renderer import Camera
from sdfest.initialization import quaternion_utils, utils


def normalize_points(points: torch.Tensor) -> torch.Tensor:
    """Normalize pointset to have zero mean.

    Normalization will be performed along second last dimension.

    Args:
        points:
            The pointsets which will be normalized,
            shape (N, M, D) or shape (M, D), N pointsets with M points of dimension D.

    Return:
        normalized_points:
            The normalized pointset, same shape as points.
        centroids:
            The means of the pointclouds used to normalize points.
            Shape (N, D) or (D,), for (N, M, D) and (M, D) inputs, respectively.
    """
    centroids = torch.mean(points, dim=-2, keepdim=True)
    normalized_points = points - centroids
    return normalized_points, centroids.squeeze()


def depth_to_pointcloud(
    depth_image: torch.Tensor,
    camera: Camera,
    normalize: bool = False,
    mask: Optional[torch.Tensor] = None,
    convention: str = "opengl",
) -> torch.Tensor:
    """Convert depth image to pointcloud.

    Args:
        depth_image: The depth image to convert to pointcloud, shape (H,W).
        camera: The camera used to lift the points.
        normalize: Whether to normalize the pointcloud with 0 centroid.
        mask:
            Only points with mask != 0 will be added to pointcloud.
            No masking will be performed if None.
        convention:
            The camera frame convention to use. One of:
                "opengl": x right, y up, z back
                "opencv": x right, y down, z forward
    Returns:
        The pointcloud in the camera frame, in OpenGL convention, shape (N,3).
    """
    fx, fy, cx, cy, _ = camera.get_pinhole_camera_parameters(0.0)

    if mask is None:
        indices = torch.nonzero(depth_image, as_tuple=True)
    else:
        indices = torch.nonzero(depth_image * mask, as_tuple=True)
    depth_values = depth_image[indices]
    points = torch.cat(
        (
            indices[1][:, None].float(),
            indices[0][:, None].float(),
            depth_values[:, None],
        ),
        dim=1,
    )

    if convention == "opengl":
        final_points = torch.empty_like(points)
        final_points[:, 0] = (points[:, 0] - cx) * points[:, 2] / fx
        final_points[:, 1] = -(points[:, 1] - cy) * points[:, 2] / fy
        final_points[:, 2] = -points[:, 2]
    elif convention == "opencv":
        final_points = torch.empty_like(points)
        final_points[:, 0] = (points[:, 0] - cx) * points[:, 2] / fx
        final_points[:, 1] = (points[:, 1] - cy) * points[:, 2] / fy
        final_points[:, 2] = points[:, 2]
    else:
        raise ValueError(f"Unsupported camera convention {convention}.")

    if normalize:
        final_points, _ = normalize_points(final_points)

    return final_points


def change_transform_camera_convention(
    in_transform: torch.Tensor, in_convention: str, out_convention: str
) -> torch.Tensor:
    """Change the camera convention for a frame A -> camera frame transform.

    Args:
        in_transform:
            Transformtion matrix(es) from coordinate frame A to in_convention camera
            frame.  Shape (...,4,4).
        in_convention:
            Camera convention for the in_transform. One of "opengl", "opencv".
        out_convention:
            Camera convention for the returned transform. One of "opengl", "opencv".
    Returns:
        Transformtion matrix(es) from coordinate frame A to out_convention camera frame.
        Same shape as in_transform.
    """
    # check whether valild convention was provided
    if in_convention not in ["opengl", "opencv"]:
        raise ValueError(f"In camera convention {in_convention} not supported.")
    if out_convention not in ["opengl", "opencv"]:
        raise ValueError(f"Out camera convention {in_convention} not supported.")

    if in_convention == out_convention:
        return in_transform
    else:
        gl2cv_transform = torch.diag(
            in_transform.new_tensor([1.0, -1.0, -1.0, 1.0])
        )  # == cv2gl_transform
        return gl2cv_transform @ in_transform


def change_position_camera_convention(
    in_position: torch.Tensor,
    in_convention: str,
    out_convention: str,
) -> tuple:
    """Change the camera convention for a position in a camera frame.

    Args:
        in_position:
            Position(s) of coordinate frame A in in_convention camera frame.
            Shape (...,3).
        in_convention:
            Camera convention for the in_position. One of "opengl", "opencv".
        out_convention:
            Camera convention for the returned transform. One of "opengl", "opencv".
    Returns:
        Position(s) of coordinate frame A in out_convention camera frame. Shape (...,3).
    """
    # check whether valild convention was provided
    if in_convention not in ["opengl", "opencv"]:
        raise ValueError(f"In camera convention {in_convention} not supported.")
    if out_convention not in ["opengl", "opencv"]:
        raise ValueError(f"Out camera convention {in_convention} not supported.")

    if in_convention == out_convention:
        return in_position
    else:
        gl2cv = in_position.new_tensor([1.0, -1.0, -1.0])  # == cv2gl
        return gl2cv * in_position


def change_orientation_camera_convention(
    in_orientation_q: torch.Tensor,
    in_convention: str,
    out_convention: str,
) -> tuple:
    """Change the camera convention for an orientation in a camera frame.

    Orientation is represented as a quaternion, that rotates points from a
    coordinate frame A to a camera frame (if those frames had the same origin).

    Args:
        in_orientation_q:
            Quaternion(s) which transforms from coordinate frame A to in_convention
            camera frame. Scalar-last convention. Shape (...,4).
        in_convention:
            Camera convention for the in_transform. One of "opengl", "opencv".
        out_convention:
            Camera convention for the returned transform. One of "opengl", "opencv".
    Returns:
        Quaternion(s) which transforms from coordinate frame A to in_convention camera
        frame. Scalar-last convention. Same shape as in_orientation_q.
    """
    # check whether valild convention was provided
    if in_convention not in ["opengl", "opencv"]:
        raise ValueError(f"In camera convention {in_convention} not supported.")
    if out_convention not in ["opengl", "opencv"]:
        raise ValueError(f"Out camera convention {in_convention} not supported.")

    if in_convention == out_convention:
        return in_orientation_q
    else:
        # rotate 180deg around x direction
        gl2cv_q = in_orientation_q.new_tensor([1.0, 0, 0, 0])  # == cv2gl
        return quaternion_utils.quaternion_multiply(gl2cv_q, in_orientation_q)


def visualize_pointset(pointset: torch.Tensor, max_points: int = 1000) -> None:
    """Visualize pointset as 3D scatter plot.

    Args:
        pointset:
            The pointset to visualize. Either shape (N,3), xyz, or shape (N,6), xyzrgb.
        max_points:
            Maximum number of points.
            If N>max_points only a random subset will be shown.
    """
    pointset_np = pointset.cpu().detach().numpy()
    fig = plt.figure()
    ax = fig.add_subplot(projection="3d")
    ax.set_box_aspect((1, 1, 1))

    if len(pointset_np) > max_points:
        indices = np.random.choice(len(pointset_np), replace=False, size=max_points)
        pointset_np = pointset_np[indices]

    if pointset_np.shape[1] == 6:
        colors = pointset_np[:, 3:]
    else:
        colors = None

    ax.scatter(pointset_np[:, 0], pointset_np[:, 1], pointset_np[:, 2], c=colors)

    ax.set_xlabel("x")
    ax.set_ylabel("y")
    ax.set_zlabel("z")

    utils.set_axes_equal(ax)
    plt.show()



"""PyTorch interface for diffferentiable renderer.

This module provides two functions:
    - render_depth: numpy-based CPU implementation
    (not recommended, only for development)
    - render_depth_gpu: CUDA implementation (fast)
"""
import math
import os

import numpy as np
import open3d as o3d
import torch
from torch.utils.cpp_extension import load
from typing import Optional, Tuple

from .simple_renderer import render_depth as _render_depth
from .simple_renderer import SDFObject

# TODO: allow JIT compilation by pip argument??
directory = os.path.dirname(__file__)
sdf_renderer_cpp = load(
    name="sdf_renderer_cpp",
    sources=[
        os.path.join(directory, "./csrc/sdf_renderer.cpp"),
        os.path.join(directory, "./csrc/sdf_renderer_cuda.cu"),
    ],
)


class Camera:
    """Pinhole camera parameters.

    This class allows conversion between different pixel conventions, i.e., pixel
    center at (0.5, 0.5) (as common in computer graphics), and (0, 0) as common in
    computer vision.
    """

    def __init__(
        self,
        width: int,
        height: int,
        fx: float,
        fy: float,
        cx: float,
        cy: float,
        s: float = 0.0,
        pixel_center: float = 0.0,
    ):
        """Initialize camera parameters.

        Note that the principal point is only fully defined in combination with
        pixel_center.

        The pixel_center defines the relation between continuous image plane
        coordinates and discrete pixel coordinates.

        A discrete image coordinate (x, y) will correspond to the continuous
        image coordinate (x + pixel_center, y + pixel_center). Normally pixel_center
        will be either 0 or 0.5. During calibration it depends on the convention
        the point features used to compute the calibration matrix.

        Note that if pixel_center == 0, the corresponding continuous coordinate
        interval for a pixel are [x-0.5, x+0.5). I.e., proper rounding has to be done
        to convert from continuous coordinate to the corresponding discrete coordinate.

        For pixel_center == 0.5, the corresponding continuous coordinate interval for a
        pixel are [x, x+1). I.e., floor is sufficient to convert from continuous
        coordinate to the corresponding discrete coordinate.

        Args:
            width: Number of pixels in horizontal direction.
            height: Number of pixels in vertical direction.
            fx: Horizontal focal length.
            fy: Vertical focal length.
            cx: Principal point x-coordinate.
            cy: Principal point y-coordinate.
            s: Skew.
            pixel_center: The center offset for the provided principal point.
        """
        # focal length
        self.fx = fx
        self.fy = fy

        # principal point
        self.cx = cx
        self.cy = cy

        self.pixel_center = pixel_center

        # skew
        self.s = s

        # image dimensions
        self.width = width
        self.height = height

    def get_o3d_pinhole_camera_parameters(self) -> o3d.camera.PinholeCameraParameters():
        """Convert camera to Open3D pinhole camera parameters.

        Open3D camera is at (0,0,0) looking along positive z axis (i.e., positive z
        values are in front of camera). Open3D expects camera with pixel_center = 0
        and does not support skew.

        Returns:
            The pinhole camera parameters.
        """
        fx, fy, cx, cy, _ = self.get_pinhole_camera_parameters(0)
        params = o3d.camera.PinholeCameraParameters()
        params.intrinsic.set_intrinsics(self.width, self.height, fx, fy, cx, cy)
        params.extrinsic = np.array(
            [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]]
        )
        return params

    def get_pinhole_camera_parameters(self, pixel_center: float) -> Tuple:
        """Convert camera to general camera parameters.

        Args:
            pixel_center:
                At which ratio of a square the pixel center should be for the resulting
                parameters. Typically 0 or 0.5. See class documentation for more info.
        Returns:
            - fx, fy: The horizontal and vertical focal length
            - cx, cy:
                The position of the principal point in continuous image plane
                coordinates considering the provided pixel center and the pixel center
                specified during the construction.
            - s: The skew.
        """
        cx_corrected = self.cx - self.pixel_center + pixel_center
        cy_corrected = self.cy - self.pixel_center + pixel_center
        return self.fx, self.fy, cx_corrected, cy_corrected, self.s


class SDFRendererFunction(torch.autograd.Function):
    """Renderer function for signed distance fields."""

    @staticmethod
    def forward(
        ctx,
        sdf: torch.Tensor,
        position: torch.Tensor,
        orientation: torch.Tensor,
        inv_scale: torch.Tensor,
        width: Optional[int] = None,
        height: Optional[int] = None,
        fov_deg: Optional[float] = None,
        threshold: Optional[float] = 0.0,
        camera: Optional[Camera] = None,
    ) -> torch.Tensor:
        """Render depth image of a 7-DOF discrete signed distance field.

        The SDF position is assumed to be in the camera frame under OpenGL convention.

        That is, camera looks along negative z-axis, y pointing upwards and x to the
        right. Note that the rendered image will still follow the classical computer
        vision convention, of first row being up in the camera frame.

        This function internally usese numpy and is very slow due to the fully serial
        implementation. This is only for testing purposes. Use the GPU version for
        practical performance.

        Camera can be specified either via camera parameter giving most
        flexbility or alternatively by providing width, height and fov_deg.

        Args:
            ctx:
                Context object to stash information.
                See https://pytorch.org/docs/stable/notes/extending.html.
            sdf:
                Discrete signed distance field with shape (M, M, M).
                Arbitrary (but uniform) resolutions are supported.
            position:
                The position of the signed distance field origin in the camera frame.
            orientation:
                The orientation of the SDF as a normalized quaternion.
            inv_scale:
                The inverted scale of the SDF. The scale of an SDF the half-width of the
                full SDF volume.
            width:
                Number of pixels in x direction. Recommended to use camera instead.
            height:
                Number of pixels in y direction. Recommended to use camera instead.
            fov_deg:
                The horizontal field of view (i.e., in x direction).
                Pixels are assumed to be square, i.e., fx=fy, computed based on width
                and fov_deg.
                Recommended to use camera instead.
            threshold:
                The distance threshold at which sphere tracing should be stopped.
                Smaller value will be more accurate, but slower and might potentially
                lead to holes in the rendering for thin structures in the SDF.
                Larger values will be faster, but will overestimate the thickness.

                Should always be positive to guarantee convergence.
            camera:
                Camera parameters (not supported right now).
        Returns:
            The rendered depth image.
        """
        if None not in [width, height, fov_deg] and camera is not None:
            raise ValueError("Either width+height+fov_dev or camera must be provided.")
        if camera is not None:
            raise NotImplementedError(
                "Only width+height+fov_dev currently supported for CPU"
            )
        # for simplicity use numpy internally
        ctx.save_for_backward(sdf, position, orientation, inv_scale)
        sdf = sdf.detach().numpy()
        position = position.detach().numpy()
        orientation = orientation.detach().numpy()
        inv_scale = inv_scale.detach().numpy()
        sdf_object = SDFObject(sdf)
        image, derivatives = _render_depth(
            sdf_object,
            width,
            height,
            fov_deg,
            "d",
            threshold,
            position,
            orientation,
            inv_scale,
        )
        ctx.derivatives = derivatives
        return torch.from_numpy(image)

    @staticmethod
    def backward(ctx, inp: torch.Tensor):
        """Compute gradients of inputs with respect to the provided gradients.

        Normally called by PyTorch as part of a call to backward() on a loss.

        Args:
            grad_depth_image:
        Returns:
            Gradients of
                discretized signed distance field, position, orientation, inverted scale
                followed by None for all the non-supported variables passed to forward.
        """
        derivatives = ctx.derivatives
        sdf, pos, quat, inv_s = ctx.saved_tensors
        g_image = inp.numpy()
        g_sdf = g_p = g_q = g_is = g_w = g_h = g_fov = g_thresh = g_camera = None
        g_sdf = torch.zeros_like(sdf)
        g_p = torch.empty_like(pos)
        g_q = torch.empty_like(quat)
        g_is = torch.empty_like(inv_s)
        g_p[0] = torch.tensor(np.sum(derivatives["x"] * g_image))
        g_p[1] = torch.tensor(np.sum(derivatives["y"] * g_image))
        g_p[2] = torch.tensor(np.sum(derivatives["z"] * g_image))
        g_q[0] = torch.tensor(np.sum(derivatives["qx"] * g_image))
        g_q[1] = torch.tensor(np.sum(derivatives["qy"] * g_image))
        g_q[2] = torch.tensor(np.sum(derivatives["qz"] * g_image))
        g_q[3] = torch.tensor(np.sum(derivatives["qw"] * g_image))
        g_is = torch.tensor(np.sum(derivatives["s_inv"] * g_image))
        if "sdf" in derivatives:
            for k, v in derivatives["sdf"].items():
                g_sdf[k] = torch.tensor(np.sum(v * g_image))
        return g_sdf, g_p, g_q, g_is, g_w, g_h, g_fov, g_thresh, g_camera


render_depth = SDFRendererFunction.apply


class SDFRendererFunctionGPU(torch.autograd.Function):
    """Renderer function for signed distance fields."""

    @staticmethod
    def forward(
        ctx,
        sdf: torch.Tensor,
        position: torch.Tensor,
        orientation: torch.Tensor,
        inv_scale: torch.Tensor,
        threshold: Optional[float] = 0.0,
        camera: Optional[Camera] = None,
    ) -> torch.Tensor:
        """Render depth image of a 7-DOF discrete signed distance field on the GPU.

        Also see render_depth_gpu for documentation.

        Args:
            ctx:
                Context object to stash information.
                See https://pytorch.org/docs/stable/notes/extending.html.
            sdf:
                Discrete signed distance field with shape (M, M, M).
                Arbitrary (but uniform) resolutions are supported.
            position:
                The position of the signed distance field origin in the camera frame.
            orientation:
                The orientation of the SDF as a normalized quaternion.
            inv_scale:
                The inverted scale of the SDF. The scale of an SDF the half-width of the
                full SDF volume.
            threshold:
                The distance threshold at which sphere tracing should be stopped.
                Smaller value will be more accurate, but slower and might potentially
                lead to holes in the rendering for thin structures in the SDF.
                Larger values will be faster, but will overestimate the thickness.

                Should always be positive to guarantee convergence.
            camera:
                Camera parameters.
        Returns:
            The rendered depth image.
        """
        fx, fy, cx, cy, _ = camera.get_pinhole_camera_parameters(0.5)
        (image,) = sdf_renderer_cpp.forward(
            sdf,
            position,
            orientation,
            inv_scale,
            camera.width,
            camera.height,
            cx,
            cy,
            fx,
            fy,
            threshold,
        )
        ctx.save_for_backward(image, sdf, position, orientation, inv_scale)
        ctx.width = camera.width
        ctx.height = camera.height
        ctx.fx = fx
        ctx.fy = fy
        ctx.cx = cx
        ctx.cy = cy
        return image

    @staticmethod
    def backward(ctx, grad_depth_image: torch.Tensor):
        """Compute gradients of inputs with respect to the provided gradients.

        Normally called by PyTorch as part of a call to backward() on a loss.

        Args:
            grad_depth_image:
        Returns:
            Gradients of
                discretized signed distance field, position, orientation, inverted scale
                followed by None for all the non-supported variables passed to forward.
        """
        g_sdf = g_p = g_q = g_is = g_thresh = g_camera = None
        g_sdf, g_p, g_q, g_is = sdf_renderer_cpp.backward(
            grad_depth_image,
            *ctx.saved_tensors,
            ctx.width,
            ctx.height,
            ctx.cx,
            ctx.cy,
            ctx.fx,
            ctx.fy
        )
        return g_sdf, g_p, g_q, g_is, g_thresh, g_camera


def render_depth_gpu(
    sdf: torch.Tensor,
    position: torch.Tensor,
    orientation: torch.Tensor,
    inv_scale: torch.Tensor,
    width: Optional[int] = None,
    height: Optional[int] = None,
    fov_deg: Optional[float] = None,
    threshold: Optional[float] = 0.0,
    camera: Optional[Camera] = None,
):
    """Render depth image of a 7-DOF discrete signed distance field on the GPU.

    The SDF position is assumed to be in the camera frame under OpenGL convention.

    That is, camera looks along negative z-axis, y pointing upwards and x to the
    right. Note that the rendered image will still follow the classical computer
    vision convention, of first row being up in the camera frame.

    Camera can be specified either via camera parameter giving most
    flexbility or alternatively by providing width, height and fov_deg.

    All provided tensors must reside on the GPU.

    Args:
        sdf:
            Discrete signed distance field with shape (M, M, M).
            Arbitrary (but uniform) resolutions are supported.
        position:
            The position of the signed distance field origin in the camera frame.
        orientation:
            The orientation of the SDF as a normalized quaternion.
        inv_scale:
            The inverted scale of the SDF. The scale of an SDF the half-width of the
            full SDF volume.
        width:
            Number of pixels in x direction. Recommended to use camera instead.
        height:
            Number of pixels in y direction. Recommended to use camera instead.
        fov_deg:
            The horizontal field of view (i.e., in x direction).
            Pixels are assumed to be square, i.e., fx=fy, computed based on width
            and fov_deg.
            Recommended to use camera instead.
        threshold:
            The distance threshold at which sphere tracing should be stopped.
            Smaller value will be more accurate, but slower and might potentially
            lead to holes in the rendering for thin structures in the SDF.
            Larger values will be faster, but will overestimate the thickness.

            Should always be positive to guarantee convergence.
        camera:
            Camera parameters.
    Returns:
        The rendered depth image.
    """
    if None not in [width, height, fov_deg] and camera is not None:
        raise ValueError("Either width+height+fov_dev or camera must be provided.")
    if camera is None:
        f = width / math.tan(fov_deg * math.pi / 180.0 / 2.0) / 2
        camera = Camera(width, height, f, f, width / 2, height / 2, pixel_center=0.5)

    return SDFRendererFunctionGPU.apply(
        sdf, position, orientation, inv_scale, threshold, camera
    )



"""Generally useful modules for pytorch."""
import torch


class View(torch.nn.Module):
    """Wrapper of torch's view method to use with nn.Sequential."""

    def __init__(self, *shape):
        """Construct the module."""
        super(View, self).__init__()
        self.shape = shape

    def forward(self, x):
        """Reshape the tensor."""
        return x.view(*self.shape)
